#!/usr/bin/env python

import errno
import os
import re
import distutils.cmd
from distutils.command.build import build
from setuptools import setup, find_packages


class WebBuildCommand(distutils.cmd.Command):
    """
    Custom command to build the web files
    """

    description = 'Build components and styles for the web pages'
    user_options = []

    @classmethod
    def generate_css_files(cls):
        try:
            # noinspection PyPackageRequirements
            from scss import Compiler
        except ImportError:
            print('pyScss module not found: {}. You will have to generate ' +
                  'the CSS files manually through python setup.py build install')
            return

        print('Building CSS files')
        base_path = path(os.path.join('platypush', 'backend', 'http', 'static', 'css'))
        input_path = path(os.path.join(base_path, 'source'))
        output_path = path(os.path.join(base_path, 'dist'))

        for root, dirs, files in os.walk(input_path, followlinks=True):
            scss_file = os.path.join(root, 'index.scss')
            if os.path.isfile(scss_file):
                css_path = os.path.split(scss_file[len(input_path):])[0][1:] + '.css'
                css_dir = os.path.join(output_path, os.path.dirname(css_path))
                css_file = os.path.join(css_dir, os.path.basename(css_path))

                os.makedirs(css_dir, exist_ok=True)
                print('\tGenerating CSS {scss} -> {css}'.format(scss=scss_file, css=css_file))

                with open(css_file, 'w') as f:
                    css_content = Compiler(output_style='compressed', search_path=[root, input_path]).compile(scss_file)
                    css_content = cls._fix_css4_vars(css_content)
                    f.write(css_content)

    @staticmethod
    def _fix_css4_vars(css):
        return re.sub(r'var\("--([^"]+)"\)', r'var(--\1)', css)

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        self.generate_css_files()


class BuildCommand(build):
    def run(self):
        build.run(self)
        self.run_command('web_build')


def path(fname=''):
    return os.path.abspath(os.path.join(os.path.dirname(__file__), fname))


def readfile(fname):
    with open(path(fname)) as f:
        return f.read()


# noinspection PyShadowingBuiltins
def pkg_files(dir):
    paths = []
    # noinspection PyShadowingNames
    for (path, dirs, files) in os.walk(dir):
        for file in files:
            paths.append(os.path.join('..', path, file))
    return paths


def create_etc_dir():
    # noinspection PyShadowingNames
    path = '/etc/platypush'
    try:
        os.makedirs(path)
    except OSError as e:
        if isinstance(e, PermissionError):
            print('WARNING: Could not create /etc/platypush')
        elif e.errno == errno.EEXIST and os.path.isdir(path):
            pass
        else:
            raise


plugins = pkg_files('platypush/plugins')
backend = pkg_files('platypush/backend')
# create_etc_dir()

setup(
    name="platypush",
    version="0.12.3",
    author="Fabio Manganiello",
    author_email="info@fabiomanganiello.com",
    description="Platypush service",
    license="MIT",
    python_requires='>= 3.5',
    keywords="home-automation iot mqtt websockets redis dashboard notificaions",
    url="https://github.com/BlackLight/platypush",
    packages=find_packages(),
    include_package_data=True,
    entry_points={
        'console_scripts': [
            'platypush=platypush:main',
            'pusher=platypush.pusher:main',
            'platydock=platypush.platydock:main',
        ],
    },
    scripts=['bin/platyvenv'],
    cmdclass={
        'web_build': WebBuildCommand,
        'build': BuildCommand,
    },
    # data_files = [
    #     ('/etc/platypush', ['platypush/config.example.yaml'])
    # ],
    long_description=readfile('README.md'),
    long_description_content_type='text/markdown',
    classifiers=[
        "Topic :: Utilities",
        "License :: OSI Approved :: MIT License",
        "Development Status :: 3 - Alpha",
    ],
    install_requires=[
        'pyyaml',
        'redis',
        'requests',
        'croniter',
        'pyScss',
        'sqlalchemy',
        'websockets',
        'websocket-client',
    ],

    extras_require={
        # Support for thread custom name
        'threadname': ['python-prctl'],
        # Support for Kafka backend and plugin
        'kafka': ['kafka-python'],
        # Support for Pushbullet backend and plugin
        'pushbullet': ['pushbullet.py @ https://github.com/rbrcsk/pushbullet.py/tarball/master'],
        # Support for HTTP backend
        'http': ['flask', 'python-dateutil', 'tz', 'frozendict', 'bcrypt'],
        # Support for uWSGI HTTP backend
        'uwsgi': ['flask', 'python-dateutil', 'tz', 'frozendict', 'uwsgi', 'bcrypt'],
        # Support for database
        'db': ['sqlalchemy'],
        # Support for MQTT backends
        'mqtt': ['paho-mqtt'],
        # Support for RSS feeds parser
        'rss': ['feedparser'],
        # Support for PDF generation
        'pdf': ['weasyprint'],
        # Support for Philips Hue plugin
        'hue': ['phue'],
        # Support for MPD/Mopidy music server plugin and backend
        'mpd': ['python-mpd2'],
        # Support for text2speech plugin
        'tts': ['mplayer'],
        # Support for Google text2speech plugin
        'google-tts': ['oauth2client', 'google-api-python-client', 'google-cloud-texttospeech'],
        # Support for OMXPlayer plugin
        'omxplayer': ['omxplayer-wrapper'],
        # Support for YouTube
        'youtube': ['youtube-dl'],
        # Support for torrents download
        'torrent': ['python-libtorrent'],
        # Support for RaspberryPi camera
        'picamera': ['picamera', 'numpy'],
        # Support for inotify file monitors
        'inotify': ['inotify'],
        # Support for Google Assistant
        'google-assistant-legacy': ['google-assistant-library'],
        'google-assistant': ['google-assistant-sdk[samples]'],
        # Support for the Google APIs
        'google': ['oauth2client', 'google-api-python-client'],
        # Support for Last.FM scrobbler plugin
        'lastfm': ['pylast'],
        # Support for custom hotword detection
        'hotword': ['snowboy'],
        # Support for real-time MIDI events
        'midi': ['rtmidi'],
        # Support for RaspberryPi GPIO
        'rpi-gpio': ['RPi.GPIO'],
        # Support for MCP3008 analog-to-digital converter plugin
        'mcp3008': ['adafruit-mcp3008'],
        # Support for smart cards detection
        'scard': ['pyscard'],
        # Support for serial port plugin
        'serial': ['pyserial'],
        # Support for ICal calendars
        'ical': ['icalendar', 'python-dateutil'],
        # Support for joystick backend
        'joystick': ['inputs'],
        # Support for Kodi plugin
        'kodi': ['kodi-json'],
        # Support for Plex plugin
        'plex': ['plexapi'],
        # Support for Chromecast plugin
        'chromecast': ['pychromecast'],
        # Support for sound devices
        'sound': ['sounddevice', 'soundfile', 'numpy'],
        # Support for web media subtitles
        'subtitles': [
            'webvtt-py',
            'python-opensubtitles @ https://github.com/agonzalezro/python-opensubtitles/tarball/master'],
        # Support for mpv player plugin
        'mpv': ['python-mpv'],
        # Support for NFC tags
        'nfc': ['nfcpy>=1.0', 'ndef'],
        # Support for enviropHAT
        'envirophat': ['envirophat'],
        # Support for GPS
        'gps': ['gps'],
        # Support for BME280 environment sensor
        'bme280': ['pimoroni-bme280'],
        # Support for LTR559 light/proximity sensor
        'ltr559': ['ltr559'],
        # Support for VL53L1X laser ranger/distance sensor
        'vl53l1x': ['smbus2', 'vl53l1x'],
        # Support for Dropbox integration
        'dropbox': ['dropbox'],
        # Support for Leap Motion backend
        'leap': ['leap-sdk @ https://github.com/BlackLight/leap-sdk-python3/tarball/master'],
        # Support for Flic buttons
        'flic': ['flic @ https://github.com/50ButtonsEach/fliclib-linux-hci/tarball/master'],
        # Support for Alexa/Echo plugin
        'alexa': ['avs @ https://github.com:BlackLight/avs/tarball/master'],
        # Support for bluetooth devices
        'bluetooth': ['pybluez', 'gattlib',
                      'pyobex @ https://github.com/BlackLight/PyOBEX/tarball/master'],
        # Support for TP-Link devices
        'tplink': ['pyHS100'],
        # Support for PWM3901 2-Dimensional Optical Flow Sensor
        'pwm3901': ['pwm3901'],
        # Support for MLX90640 thermal camera
        'mlx90640': ['Pillow'],
        # Support for machine learning and CV plugin
        'cv': ['cv2', 'numpy'],
        # Support for the generation of HTML documentation from docstring
        'htmldoc': ['docutils'],
        # Support for Node-RED integration
        'nodered': ['pynodered'],
        # Support for Todoist integration
        'todoist': ['todoist-python'],
        # Support for Trello integration
        'trello': ['py-trello'],
        # Support for Google Pub/Sub
        'google-pubsub': ['google-cloud-pubsub'],
        # Support for keyboard/mouse plugin
        'inputs': ['pyuserinput'],
        # Support for Buienradar weather forecast
        'buienradar': ['buienradar'],
        # Support for Telegram integration
        'telegram': ['python-telegram-bot'],
        # Support for Arduino integration
        'arduino': ['pyserial', 'pyfirmata2'],
        # Support for CUPS printers management
        'cups': ['pycups'],
        # Support for Graphite integration
        'graphite': ['graphyte'],
        # Support for CPU and memory monitoring and info
        'sys': ['py-cpuinfo', 'psutil'],
        # Support for nmap integration
        'nmap': ['python-nmap'],
        # Support for zigbee2mqtt
        'zigbee': ['paho-mqtt'],
        # Support for Z-Wave
        'zwave': ['python-openzwave'],
        # Support for Mozilla DeepSpeech speech-to-text engine
        'deepspeech': ['deepspeech', 'numpy','sounddevice'],
        # Support for PicoVoice hotword detection engine
        'picovoice-hotword': ['pvporcupine'],
        # Support for PicoVoice speech-to-text engine
        'picovoice-speech': ['pvcheetah @ git+https://github.com/BlackLight/cheetah'],
        # Support for OTP (One-Time Password) generation
        'otp': ['pyotp'],
        # Support for Linode integration
        'linode': ['linode_api4'],
        # Support for QR codes
        'qrcode': ['numpy','qrcode[pil]', 'Pillow', 'pyzbar'],
        # Support for Tensorflow
        'tensorflow': ['numpy', 'tensorflow>=2.0', 'keras', 'pandas'],
    },
)



from platypush.message.event import Event


class PushbulletEvent(Event):
    """
    PushBullet event object.

    If you have configured the PushBullet backend with your account token,
    and enabled notification mirroring on the PushBullet app on your mobile
    devices, then the backend will trigger a PushbulletEvent whenever
    a new notiification hits your mobile, and you can react to that event
    through hooks that can, for example, log your notifications on a database,
    display them on a dashboard, let the built-in text-to-speech plugin read
    them out loud to you if they match the package name of your news app,
    display them on your smart watch if they are pictures, and so on.
    """

    def __init__(self, *args, **kwargs):
        """ Platypush supports by default the PushBullet notification mirror
            format, https://docs.pushbullet.com/#mirrored-notifications """

        if 'type' in kwargs:
            # Prevent name clash with event type attribute
            kwargs['push_type'] = kwargs.pop('type')

        super().__init__(*args, **kwargs)


# vim:sw=4:ts=4:et:




"""
.. moduleauthor:: Fabio Manganiello <blacklight86@gmail.com>
.. license: MIT
"""

import os
import threading

from platypush.backend.assistant import AssistantBackend
from platypush.context import get_plugin
from platypush.message.event.assistant import HotwordDetectedEvent


class AssistantSnowboyBackend(AssistantBackend):
    """
    Backend for detecting custom voice hotwords through Snowboy.  The purpose of
    this component is only to detect the hotword specified in your Snowboy voice
    model. If you want to trigger proper assistant conversations or custom
    speech recognition, you should create a hook in your configuration on
    HotwordDetectedEvent to trigger the conversation on whichever assistant
    plugin you're using (Google, Alexa...)

    Triggers:

        * :class:`platypush.message.event.assistant.HotwordDetectedEvent` \
            whenever the hotword has been detected

    Requires:

        * **snowboy** (``pip install snowboy``)

    Manual installation for snowboy and its Python bindings if the command above fails::

        $ [sudo] apt-get install libatlas-base-dev swig
        $ [sudo] pip install pyaudio
        $ git clone https://github.com/Kitt-AI/snowboy
        $ cd snowboy/swig/Python3
        $ make
        $ cd ../..
        $ python3 setup.py build
        $ [sudo] python setup.py install

    You will also need a voice model for the hotword detection. You can find
    some under the ``resources/models`` directory of the Snowboy repository,
    or train/download other models from https://snowboy.kitt.ai.
    """

    def __init__(self, models, audio_gain=1.0, **kwargs):
        """
        :param models: Map (name -> configuration) of voice models to be used by
            the assistant.  See https://snowboy.kitt.ai/ for training/downloading
            models. Sample format::

                ok_google:    # Hotword model name
                    voice_model_file: /path/models/OK Google.pmdl  # Voice model file location
                    sensitivity: 0.5            # Model sensitivity, between 0 and 1 (default: 0.5)
                    assistant_plugin: assistant.google.pushtotalk  # When the hotword is detected trigger the Google
                                                                   # push-to-talk assistant plugin (optional)
                    assistant_language: en-US   # The assistant will conversate in English when this hotword is
                        detected (optional)
                    detect_sound: /path/to/bell.wav   # Sound file to be played when the hotword is detected (optional)

                ciao_google:  # Hotword model name
                    voice_model_file: /path/models/Ciao Google.pmdl  # Voice model file location
                    sensitivity: 0.5            # Model sensitivity, between 0 and 1 (default: 0.5)
                    assistant_plugin: assistant.google.pushtotalk    # When the hotword is detected trigger the Google
                                                                     # push-to-talk assistant plugin (optional)
                    assistant_language: it-IT   # The assistant will conversate in Italian when this hotword is
                                                # detected (optional)
                    detect_sound: /path/to/bell.wav   # Sound file to be played when the hotword is detected (optional)

        :type models: dict

        :param audio_gain: Audio gain, between 0 and 1. Default: 1
        :type audio_gain: float
        """

        try:
            import snowboydecoder
        except ImportError:
            import snowboy.snowboydecoder as snowboydecoder

        super().__init__(**kwargs)

        self.models = {}
        self._init_models(models)
        self.audio_gain = audio_gain

        self.detector = snowboydecoder.HotwordDetector(
            [model['voice_model_file'] for model in self.models.values()],
            sensitivity=[model['sensitivity'] for model in self.models.values()],
            audio_gain=self.audio_gain)

        self.logger.info('Initialized Snowboy hotword detection with {} voice model configurations'.
                         format(len(self.models)))

    def _init_models(self, models):
        if not models:
            raise AttributeError('Please specify at least one voice model')

        self.models = {}
        for name, conf in models.items():
            if name in self.models:
                raise AttributeError('Duplicate model key {}'.format(name))

            model_file = conf.get('voice_model_file')
            detect_sound = conf.get('detect_sound')

            if not model_file:
                raise AttributeError('No voice_model_file specified for model {}'.format(name))

            model_file = os.path.abspath(os.path.expanduser(model_file))
            assistant_plugin_name = conf.get('assistant_plugin')

            if detect_sound:
                detect_sound = os.path.abspath(os.path.expanduser(detect_sound))

            if not os.path.isfile(model_file):
                raise FileNotFoundError('Voice model file {} does not exist or it not a regular file'.
                                        format(model_file))

            self.models[name] = {
                'voice_model_file': model_file,
                'sensitivity': conf.get('sensitivity', 0.5),
                'detect_sound': detect_sound,
                'assistant_plugin': get_plugin(assistant_plugin_name) if assistant_plugin_name else None,
                'assistant_language': conf.get('assistant_language'),
                'tts_plugin': conf.get('tts_plugin'),
                'tts_args': conf.get('tts_args', {}),
            }

    def hotword_detected(self, hotword):
        """
        Callback called on hotword detection
        """
        try:
            import snowboydecoder
        except ImportError:
            import snowboy.snowboydecoder as snowboydecoder

        def sound_thread(sound):
            snowboydecoder.play_audio_file(sound)

        def callback():
            if not self.is_detecting():
                self.logger.info('Hotword detected but assistant response currently paused')
                return

            self.bus.post(HotwordDetectedEvent(hotword=hotword))
            model = self.models[hotword]

            detect_sound = model.get('detect_sound')
            assistant_plugin = model.get('assistant_plugin')
            assistant_language = model.get('assistant_language')
            tts_plugin = model.get('tts_plugin')
            tts_args = model.get('tts_args')

            if detect_sound:
                threading.Thread(target=sound_thread, args=(detect_sound,)).start()

            if assistant_plugin:
                assistant_plugin.start_conversation(language=assistant_language, tts_plugin=tts_plugin,
                                                    tts_args=tts_args)

        return callback

    def on_stop(self):
        if self.detector:
            self.detector.terminate()
            self.detector = None

    def run(self):
        super().run()
        self.detector.start(detected_callback=[
            self.hotword_detected(hotword)
            for hotword in self.models.keys()
        ])


# vim:sw=4:ts=4:et:



from platypush.message.response import Response


class SpeechDetectedResponse(Response):
    def __init__(self, *args, speech: str, **kwargs):
        super().__init__(*args, output={
            'speech': speech
        }, **kwargs)


# vim:sw=4:ts=4:et:



from platypush.backend.sensor import SensorBackend


class SensorAccelerometerBackend(SensorBackend):
    """
    Backend to poll position information from an accelerometer sensor.

    Requires:

        * ``Adafruit_Python_GPIO`` (``pip install Adafruit_Python_GPIO``)
        * The :mod:`platypush.plugins.gpio.sensor.accelerometer` plugin configured
    """

    def __init__(self, **kwargs):
        super().__init__(plugin='gpio.sensor.accelerometer', **kwargs)


# vim:sw=4:ts=4:et:



from platypush.message.event import Event

class WebhookEvent(Event):
    """
    Event triggered when a custom webhook is called.
    """

    def __init__(self, hook, method, data=None, args=None, *argv, **kwargs):
        """
        :param hook: Name of the invoked web hook, from http://host:port/hook/<hook>
        :type hook: str

        :param method: HTTP method (in uppercase)
        :type method: str

        :param data: Extra data passed over POST/PUT/DELETE
        :type data: str or dict/list from JSON

        :param args: Extra query string arguments
        :type args: dict
        """

        super().__init__(hook=hook, method=method, data=data,
                         args=args or {}, *argv, **kwargs)


# vim:sw=4:ts=4:et:




import json

from threading import Event
from typing import List
from websocket import WebSocketApp

from platypush.backend import Backend
from platypush.context import get_plugin
from platypush.message.event.trello import MoveCardEvent, NewCardEvent, ArchivedCardEvent, \
        UnarchivedCardEvent

from platypush.plugins.trello import TrelloPlugin


class TrelloBackend(Backend):
    """
    This backend listens for events on a remote Trello account through websocket interface..
    Note that the Trello websocket interface
    `is not officially supported <https://community.developer.atlassian.com/t/websocket-interface/34586/4>`_
    and it requires a different token from the one you use for the Trello API (and for the Trello plugin).
    To get the websocket token:

        1. Open https://trello.com in your browser.
        2. Open the developer tools (F12), go to the Network tab, select 'Websocket' or 'WS' in the filter bar
           and refresh the page.
        3. You should see an entry in the format ``wss://trello.com/1/Session/socket?token=<token>``.
        4. Copy the ``<token>`` in the configuration of this backend.

    Requires:

        * **websocket-client** (``pip install websocket-client``)
        * The :class:`platypush.plugins.trello.TrelloPlugin` configured.

    Triggers:

        * :class:`platypush.message.event.trello.NewCardEvent` when a card is created.
        * :class:`platypush.message.event.MoveCardEvent` when a card is moved.
        * :class:`platypush.message.event.ArchivedCardEvent` when a card is archived/closed.
        * :class:`platypush.message.event.UnarchivedCardEvent` when a card is un-archived/opened.

    """

    _websocket_url_base = 'wss://trello.com/1/Session/socket?token={token}'

    def __init__(self, boards: List[str], token: str, **kwargs):
        """
        :param boards: List of boards to subscribe, by ID or name.
        :param token: Trello web client API token.
        """

        super().__init__(**kwargs)
        self._plugin: TrelloPlugin = get_plugin('trello')
        self.token = token
        self._req_id = 0
        self._boards_by_id = {}
        self._boards_by_name = {}

        for b in boards:
            b = self._plugin.get_board(b).board
            self._boards_by_id[b.id] = b
            self._boards_by_name[b.name] = b

        self.url = self._websocket_url_base.format(token=self.token)
        self._connected = Event()

        self._items = {}
        self._event_handled = False

    def _initialize_connection(self, ws: WebSocketApp):
        for board_id in self._boards_by_id.keys():
            self._send(ws, {
                'type': 'subscribe',
                'modelType': 'Board',
                'idModel': board_id,
                'tags': ['clientActions', 'updates'],
                'invitationTokens': [],
            })

        self.logger.info('Trello boards subscribed')

    def _on_msg(self):
        # noinspection PyUnusedLocal
        def hndl(ws: WebSocketApp, msg):
            if not msg:
                # Reply back with an empty message when the server sends an empty message
                ws.send('')
                return

            # noinspection PyBroadException
            try:
                msg = json.loads(msg)
            except:
                self.logger.warning('Received invalid JSON message from Trello: {}'.format(msg))
                return

            if 'error' in msg:
                self.logger.warning('Trello error: {}'.format(msg['error']))
                return

            if msg.get('reqid') == 0:
                self.logger.debug('Ping response received, subscribing boards')
                self._initialize_connection(ws)
                return

            notify = msg.get('notify')
            if not notify:
                return

            if notify['event'] != 'updateModels' or notify['typeName'] != 'Action':
                return

            for delta in notify['deltas']:
                args = {
                    'card_id': delta['data']['card']['id'],
                    'card_name': delta['data']['card']['name'],
                    'list_id': (delta['data'].get('list') or delta['data'].get('listAfter', {})).get('id'),
                    'list_name': (delta['data'].get('list') or delta['data'].get('listAfter', {})).get('name'),
                    'board_id': delta['data']['board']['id'],
                    'board_name': delta['data']['board']['name'],
                    'closed': delta.get('closed'),
                    'member_id': delta['memberCreator']['id'],
                    'member_username': delta['memberCreator']['username'],
                    'member_fullname': delta['memberCreator']['fullName'],
                    'date': delta['date'],
                }

                if delta.get('type') == 'createCard':
                    self.bus.post(NewCardEvent(**args))
                elif delta.get('type') == 'updateCard':
                    if 'listBefore' in delta['data']:
                        args.update({
                            'old_list_id': delta['data']['listBefore']['id'],
                            'old_list_name': delta['data']['listBefore']['name'],
                        })

                        self.bus.post(MoveCardEvent(**args))
                    elif 'closed' in delta['data'].get('old', {}):
                        cls = UnarchivedCardEvent if delta['data']['old']['closed'] else ArchivedCardEvent
                        self.bus.post(cls(**args))

        return hndl

    def _on_error(self):
        # noinspection PyUnusedLocal
        def hndl(ws: WebSocketApp, error):
            self.logger.warning('Trello websocket error: {}'.format(error))

        return hndl

    def _on_close(self):
        # noinspection PyUnusedLocal
        def hndl(ws: WebSocketApp):
            self.logger.warning('Trello websocket connection closed')
            self._connected.clear()
            self._req_id = 0

            while True:
                try:
                    self._connect()
                    self._connected.wait(timeout=20)
                    break
                except TimeoutError:
                    continue

        return hndl

    def _on_open(self):
        # noinspection PyUnusedLocal
        def hndl(ws: WebSocketApp):
            self._connected.set()
            self._send(ws, {'type': 'ping'})
            self.logger.info('Trello websocket connected')

        return hndl

    def _send(self, ws: WebSocketApp, msg: dict):
        msg['reqid'] = self._req_id
        ws.send(json.dumps(msg))
        self._req_id += 1

    def _connect(self) -> WebSocketApp:
        return WebSocketApp(self.url,
                            on_open=self._on_open(),
                            on_message=self._on_msg(),
                            on_error=self._on_error(),
                            on_close=self._on_close())

    def run(self):
        super().run()
        self.logger.info('Started Todoist backend')
        ws = self._connect()
        ws.run_forever()


# vim:sw=4:ts=4:et:



from .context import platypush, config_file, TestTimeoutException

import json
import logging
import requests
import sys
import time
import unittest

from threading import Thread

from platypush import Daemon
from platypush.config import Config
from platypush.message import Message
from platypush.message.response import Response
from platypush.utils import set_timeout, clear_timeout

class TestHttp(unittest.TestCase):
    """ Tests the full flow of a request/response on the HTTP backend.
        Runs a remote command over HTTP via shell.exec plugin and gets the output """

    timeout = 10
    sleep_secs = 10

    def setUp(self):
        logging.basicConfig(level=logging.INFO, stream=sys.stdout)
        backends = Config.get_backends()
        self.assertTrue('http' in backends)

    def test_request_exec_flow(self):
        self.start_daemon()
        logging.info('Sleeping {} seconds while waiting for the daemon to start up'.format(self.sleep_secs))
        time.sleep(self.sleep_secs)
        self.send_request()

    def start_daemon(self):
        def _f():
            self.receiver = Daemon(config_file=config_file)
            self.receiver.start()

        Thread(target=_f).start()

    def on_timeout(self, msg):
        def _f(): raise TestTimeoutException(msg)
        return _f

    def send_request(self):
        set_timeout(seconds=self.timeout,
                    on_timeout=self.on_timeout('Receiver response timed out'))

        response = requests.post(
            u'http://localhost:8123/execute',
            json  = {
                'type': 'request',
                'target': Config.get('device_id'),
                'action': 'shell.exec',
                'args': { 'cmd':'echo ping' }
            }
        )

        clear_timeout()

        response = Message.build(response.json())
        self.assertTrue(isinstance(response, Response))
        self.assertEqual(response.output.strip(), 'ping')
        self.receiver.stop_app()


if __name__ == '__main__':
    unittest.main()


# vim:sw=4:ts=4:et:



from platypush.plugins import action
from platypush.plugins.gpio.sensor import GpioSensorPlugin


class GpioSensorBme280Plugin(GpioSensorPlugin):
    """
    Plugin to interact with a `BME280 <https://shop.pimoroni.com/products/bme280-breakout>`_ environment sensor for
    temperature, humidity and pressure measurements over I2C interface

    Requires:

        * ``pimoroni-bme280`` (``pip install pimoroni-bme280``)
    """

    def __init__(self, port=1, **kwargs):
        """
        :param port: I2C port. 0 = /dev/i2c-0 (port I2C0), 1 = /dev/i2c-1 (port I2C1)
        """

        super().__init__(**kwargs)
        self.port = port
        self._bus = None
        self._device = None

    # noinspection PyPackageRequirements
    # noinspection PyUnresolvedReferences
    def _get_device(self):
        if self._device:
            return self._device

        from smbus import SMBus
        from bme280 import BME280

        self._bus = SMBus(self.port)
        self._device = BME280(i2c_dev=self._bus)
        return self._device

    @action
    def get_measurement(self):
        """
        :returns: dict. Example:

        .. code-block:: python

           output = {
               "temperature": 21.0,   # Celsius
               "pressure": 101555.08, # Pascals
               "humidity": 23.543,    # percentage
               "altitude": 15.703     # meters
           }

        """

        device = self._get_device()
        return {
            'temperature': device.get_temperature(),
            'pressure': device.get_pressure()*100,
            'humidity': device.get_humidity(),
            'altitude': device.get_altitude(),
        }


# vim:sw=4:ts=4:et:



from typing import Union, Dict, List

from pyHS100 import SmartDevice, SmartPlug, SmartBulb, SmartStrip, Discover

from platypush.plugins import action
from platypush.plugins.switch import SwitchPlugin


class SwitchTplinkPlugin(SwitchPlugin):
    """
    Plugin to interact with TP-Link smart switches/plugs like the HS100
    (https://www.tp-link.com/us/products/details/cat-5516_HS100.html).

    Requires:

        * **pyHS100** (``pip install pyHS100``)

    """

    _ip_to_dev = {}
    _alias_to_dev = {}

    def __init__(self, plugs: Union[Dict[str, str], List[str]] = None, bulbs: Union[Dict[str, str], List[str]] = None,
                 strips: Union[Dict[str, str], List[str]] = None, **kwargs):
        """
        :param plugs: Optional list of IP addresses or name->address mapping if you have a static list of
            TpLink plugs and you want to save on the scan time.
        :param bulbs: Optional list of IP addresses or name->address mapping if you have a static list of
            TpLink bulbs and you want to save on the scan time.
        :param strips: Optional list of IP addresses or name->address mapping if you have a static list of
            TpLink strips and you want to save on the scan time.
        """
        super().__init__(**kwargs)
        self._ip_to_dev = {}
        self._alias_to_dev = {}
        self._static_devices = {}

        if isinstance(plugs, list):
            plugs = {addr: addr for addr in plugs}
        if isinstance(bulbs, list):
            bulbs = {addr: addr for addr in bulbs}
        if isinstance(strips, list):
            strips = {addr: addr for addr in strips}

        for name, addr in (plugs or {}).items():
            self._static_devices[addr] = {
                'name': name,
                'type': SmartPlug,
            }

        for name, addr in (bulbs or {}).items():
            self._static_devices[addr] = {
                'name': name,
                'type': SmartBulb,
            }

        for name, addr in (strips or {}).items():
            self._static_devices[addr] = {
                'name': name,
                'type': SmartStrip,
            }

        self._update_devices()

    def _update_devices(self, devices: Dict[str, SmartDevice] = None):
        for (addr, info) in self._static_devices.items():
            dev = info['type'](addr)
            self._ip_to_dev[addr] = dev
            self._alias_to_dev[info.get('name', dev.alias)] = dev

        for (ip, dev) in (devices or {}).items():
            self._ip_to_dev[ip] = dev
            self._alias_to_dev[dev.alias] = dev

    def _scan(self):
        devices = Discover.discover()
        self._update_devices(devices)
        return devices

    def _get_device(self, device, use_cache=True):
        if not use_cache:
            self._scan()

        if device in self._ip_to_dev:
            return self._ip_to_dev[device]

        if device in self._alias_to_dev:
            return self._alias_to_dev[device]

        if use_cache:
            return self._get_device(device, use_cache=False)
        else:
            raise RuntimeError('Device {} not found'.format(device))

    @action
    def on(self, device, **kwargs):
        """
        Turn on a device

        :param device: Device IP, hostname or alias
        :type device: str
        """

        device = self._get_device(device)
        device.turn_on()
        return self.status(device)

    @action
    def off(self, device, **kwargs):
        """
        Turn off a device

        :param device: Device IP, hostname or alias
        :type device: str
        """

        device = self._get_device(device)
        device.turn_off()
        return self.status(device)

    @action
    def toggle(self, device, **kwargs):
        """
        Toggle the state of a device (on/off)

        :param device: Device IP, hostname or alias
        :type device: str
        """

        device = self._get_device(device)

        if device.is_on:
            device.turn_off()
        else:
            device.turn_on()

        return {
            'current_consumption': device.current_consumption(),
            'id': device.host,
            'ip': device.host,
            'host': device.host,
            'hw_info': device.hw_info,
            'name': device.alias,
            'on': device.is_on,
        }

    @property
    def devices(self):
        return [
            {
                'current_consumption': dev.current_consumption(),
                'id': ip,
                'ip': ip,
                'host': dev.host,
                'hw_info': dev.hw_info,
                'name': dev.alias,
                'on': dev.is_on,
            } for (ip, dev) in self._scan().items()
        ]


# vim:sw=4:ts=4:et:



import json
from typing import Optional

from platypush.backend.mqtt import MqttBackend
from platypush.context import get_plugin
from platypush.message.event.zigbee.mqtt import ZigbeeMqttOnlineEvent, ZigbeeMqttOfflineEvent, \
        ZigbeeMqttDevicePropertySetEvent, ZigbeeMqttDevicePairingEvent, ZigbeeMqttDeviceConnectedEvent, \
        ZigbeeMqttDeviceBannedEvent, ZigbeeMqttDeviceRemovedEvent, ZigbeeMqttDeviceRemovedFailedEvent, \
        ZigbeeMqttDeviceWhitelistedEvent, ZigbeeMqttDeviceRenamedEvent, ZigbeeMqttDeviceBindEvent, \
        ZigbeeMqttDeviceUnbindEvent, ZigbeeMqttGroupAddedEvent, ZigbeeMqttGroupAddedFailedEvent, \
        ZigbeeMqttGroupRemovedEvent, ZigbeeMqttGroupRemovedFailedEvent, ZigbeeMqttGroupRemoveAllEvent, \
        ZigbeeMqttGroupRemoveAllFailedEvent, ZigbeeMqttErrorEvent


class ZigbeeMqttBackend(MqttBackend):
    """
    Listen for events on a zigbee2mqtt service.

    Triggers:

        * :class:`platypush.message.event.zigbee.mqtt.ZigbeeMqttOnlineEvent` when the service comes online.
        * :class:`platypush.message.event.zigbee.mqtt.ZigbeeMqttOfflineEvent` when the service goes offline.
        * :class:`platypush.message.event.zigbee.mqtt.ZigbeeMqttDevicePropertySetEvent` when the properties of a
          connected device change.
        * :class:`platypush.message.event.zigbee.mqtt.ZigbeeMqttDevicePairingEvent` when a device is pairing.
        * :class:`platypush.message.event.zigbee.mqtt.ZigbeeMqttDeviceConnectedEvent` when a device connects
          to the network.
        * :class:`platypush.message.event.zigbee.mqtt.ZigbeeMqttDeviceBannedEvent` when a device is banned
          from the network.
        * :class:`platypush.message.event.zigbee.mqtt.ZigbeeMqttDeviceRemovedEvent` when a device is removed
          from the network.
        * :class:`platypush.message.event.zigbee.mqtt.ZigbeeMqttDeviceRemovedFailedEvent` when a request to
          remove a device from the network fails.
        * :class:`platypush.message.event.zigbee.mqtt.ZigbeeMqttDeviceWhitelistedEvent` when a device is
          whitelisted on the network.
        * :class:`platypush.message.event.zigbee.mqtt.ZigbeeMqttDeviceRenamedEvent` when a device is
          renamed on the network.
        * :class:`platypush.message.event.zigbee.mqtt.ZigbeeMqttDeviceBindEvent` when a device bind event
          occurs.
        * :class:`platypush.message.event.zigbee.mqtt.ZigbeeMqttDeviceUnbindEvent` when a device unbind event
          occurs.
        * :class:`platypush.message.event.zigbee.mqtt.ZigbeeMqttGroupAddedEvent` when a group is added.
        * :class:`platypush.message.event.zigbee.mqtt.ZigbeeMqttGroupAddedFailedEvent` when a request to
          add a new group fails.
        * :class:`platypush.message.event.zigbee.mqtt.ZigbeeMqttGroupRemovedEvent` when a group is removed.
        * :class:`platypush.message.event.zigbee.mqtt.ZigbeeMqttGroupRemovedFailedEvent` when a request to
          remove a group fails.
        * :class:`platypush.message.event.zigbee.mqtt.ZigbeeMqttGroupRemoveAllEvent` when all the devices
          are removed from a group.
        * :class:`platypush.message.event.zigbee.mqtt.ZigbeeMqttGroupRemoveAllFailedEvent` when a request to
          remove all the devices from a group fails.
        * :class:`platypush.message.event.zigbee.mqtt.ZigbeeMqttErrorEvent` when an internal error occurs
          on the zigbee2mqtt service.

    Requires:

        * **paho-mqtt** (``pip install paho-mqtt``)
        * The :class:`platypush.plugins.zigbee.mqtt.ZigbeeMqttPlugin` plugin configured.

    """

    def __init__(self, host: Optional[str] = None, port: Optional[int] = None, base_topic='zigbee2mqtt',
                 tls_cafile: Optional[str] = None, tls_certfile: Optional[str] = None,
                 tls_keyfile: Optional[str] = None, tls_version: Optional[str] = None,
                 tls_ciphers: Optional[str] = None, username: Optional[str] = None,
                 password: Optional[str] = None, *args, **kwargs):
        """
        :param host: MQTT broker host (default: host configured on the ``zigbee.mqtt`` plugin).
        :param port: MQTT broker port (default: 1883).
        :param base_topic: Prefix of the topics published by zigbee2mqtt (default: '``zigbee2mqtt``').
        :param tls_cafile: If TLS/SSL is enabled on the MQTT server and the certificate requires a certificate authority
            to authenticate it, `ssl_cafile` will point to the provided ca.crt file (default: None)
        :param tls_certfile: If TLS/SSL is enabled on the MQTT server and a client certificate it required, specify it
            here (default: None)
        :param tls_keyfile: If TLS/SSL is enabled on the MQTT server and a client certificate key it required,
            specify it here (default: None) :type tls_keyfile: str
        :param tls_version: If TLS/SSL is enabled on the MQTT server and it requires a certain TLS version, specify it
            here (default: None)
        :param tls_ciphers: If TLS/SSL is enabled on the MQTT server and an explicit list of supported ciphers is
            required, specify it here (default: None)
        :param username: Specify it if the MQTT server requires authentication (default: None)
        :param password: Specify it if the MQTT server requires authentication (default: None)
        """

        if host:
            self.base_topic = base_topic
            listeners = [{
                'host': host,
                'port': port or self._default_mqtt_port,
                'tls_cafile': tls_cafile,
                'tls_certfile': tls_certfile,
                'tls_ciphers': tls_ciphers,
                'tls_keyfile': tls_keyfile,
                'tls_version': tls_version,
                'username': username,
                'password': password,
                'topics': [
                    base_topic + '/' + topic
                    for topic in ['bridge/state', 'bridge/log']
                ],
            }]
        else:
            plugin = get_plugin('zigbee.mqtt')
            self.base_topic = plugin.base_topic
            listeners = [{
                'host': plugin.host,
                'port': plugin.port or self._default_mqtt_port,
                'tls_cafile': plugin.tls_cafile,
                'tls_certfile': plugin.tls_certfile,
                'tls_ciphers': plugin.tls_ciphers,
                'username': plugin.username,
                'password': plugin.password,
                'topics': [
                    plugin.base_topic + '/' + topic
                    for topic in ['bridge/state', 'bridge/log']
                ],
            }]

        super().__init__(subscribe_default_topic=False, listeners=listeners, *args, **kwargs)
        self._devices = {}

    def _process_state_message(self, client, msg):
        if msg == 'online':
            evt = ZigbeeMqttOnlineEvent
            self._refresh_devices(client)
        elif msg == 'offline':
            evt = ZigbeeMqttOfflineEvent
            self.logger.warning('zigbee2mqtt service is offline')
        else:
            return

        # noinspection PyProtectedMember
        self.bus.post(evt(host=client._host, port=client._port))

    def _refresh_devices(self, client):
        client.publish(self.base_topic + '/' + 'bridge/config/devices/get')

    def _process_log_message(self, client, msg):
        msg_type = msg.get('type')
        msg = msg.get('message')
        # noinspection PyProtectedMember
        args = {'host': client._host, 'port': client._port}

        if msg_type == 'devices':
            devices = {}
            for dev in (msg or []):
                devices[dev['friendly_name']] = dev
                client.subscribe(self.base_topic + '/' + dev['friendly_name'])

            self._devices = devices
        elif msg_type == 'pairing':
            self.bus.post(ZigbeeMqttDevicePairingEvent(device=msg, **args))
        elif msg_type == 'device_connected':
            self.bus.post(ZigbeeMqttDeviceConnectedEvent(device=msg, **args))
            self._refresh_devices(client)
        elif msg_type in ['device_ban', 'device_banned']:
            self.bus.post(ZigbeeMqttDeviceBannedEvent(device=msg, **args))
        elif msg_type in ['device_removed', 'device_force_removed']:
            force = msg_type == 'device_force_removed'
            self.bus.post(ZigbeeMqttDeviceRemovedEvent(device=msg, force=force, **args))
        elif msg_type in ['device_removed_failed', 'device_force_removed_failed']:
            force = msg_type == 'device_force_removed_failed'
            self.bus.post(ZigbeeMqttDeviceRemovedFailedEvent(device=msg, force=force, **args))
        elif msg_type == 'device_whitelisted':
            self.bus.post(ZigbeeMqttDeviceWhitelistedEvent(device=msg, **args))
        elif msg_type == 'device_renamed':
            self.bus.post(ZigbeeMqttDeviceRenamedEvent(device=msg, **args))
            self._refresh_devices(client)
        elif msg_type == 'device_bind':
            self.bus.post(ZigbeeMqttDeviceBindEvent(device=msg, **args))
        elif msg_type == 'device_unbind':
            self.bus.post(ZigbeeMqttDeviceUnbindEvent(device=msg, **args))
        elif msg_type == 'device_group_add':
            self.bus.post(ZigbeeMqttGroupAddedEvent(group=msg, **args))
        elif msg_type == 'device_group_add_failed':
            self.bus.post(ZigbeeMqttGroupAddedFailedEvent(group=msg, **args))
        elif msg_type == 'device_group_remove':
            self.bus.post(ZigbeeMqttGroupRemovedEvent(group=msg, **args))
        elif msg_type == 'device_group_remove_failed':
            self.bus.post(ZigbeeMqttGroupRemovedFailedEvent(group=msg, **args))
        elif msg_type == 'device_group_remove_all':
            self.bus.post(ZigbeeMqttGroupRemoveAllEvent(group=msg, **args))
        elif msg_type == 'device_group_remove_all_failed':
            self.bus.post(ZigbeeMqttGroupRemoveAllFailedEvent(group=msg, **args))
        elif msg_type == 'zigbee_publish_error':
            self.logger.warning('zigbee2mqtt internal error: {}'.format(msg))
            self.bus.post(ZigbeeMqttErrorEvent(error=msg, **args))

    def on_mqtt_message(self):
        def handler(client, _, msg):
            topic = msg.topic[len(self.base_topic)+1:]
            data = msg.payload.decode()

            # noinspection PyBroadException
            try:
                data = json.loads(data)
            except:
                pass

            if topic == 'bridge/state':
                self._process_state_message(client, data)
            elif topic == 'bridge/log':
                self._process_log_message(client, data)
            else:
                # noinspection PyProtectedMember
                self.bus.post(ZigbeeMqttDevicePropertySetEvent(host=client._host, port=client._port,
                                                               device=topic, properties=data))

        return handler


# vim:sw=4:ts=4:et:



from platypush.backend.stt import SttBackend


class SttDeepspeechBackend(SttBackend):
    """
    Backend for the Mozilla Deepspeech speech-to-text engine plugin. Set this plugin to ``enabled`` if you
    want to run the speech-to-text engine continuously instead of programmatically using
    ``start_detection`` and ``stop_detection``.

    Requires:

        - The :class:`platypush.plugins.stt.deepspeech.SttDeepspeechPlugin` plugin configured and its dependencies
          installed, as well as the language model files.

    """

    def __init__(self, *args, **kwargs):
        super().__init__('stt.deepspeech', *args, **kwargs)


# vim:sw=4:ts=4:et:



import json
import os
import threading
from typing import Optional

from platypush.backend import Backend
from platypush.context import get_plugin
from platypush.message import Message
from platypush.message.event.mqtt import MQTTMessageEvent
from platypush.message.request import Request
from platypush.utils import set_thread_name


class MqttBackend(Backend):
    """
    Backend that reads messages from a configured MQTT topic (default:
    ``platypush_bus_mq/<device_id>``) and posts them to the application bus.

    Triggers:

        * :class:`platypush.message.event.mqtt.MQTTMessageEvent` when a new
            message is received on one of the custom listeners

    Requires:

        * **paho-mqtt** (``pip install paho-mqtt``)
    """

    _default_mqtt_port = 1883

    def __init__(self, host: Optional[str] = None, port: int = _default_mqtt_port,
                 topic='platypush_bus_mq', subscribe_default_topic: bool = True,
                 tls_cafile: Optional[str] = None, tls_certfile: Optional[str] = None,
                 tls_keyfile: Optional[str] = None, tls_version: Optional[str] = None,
                 tls_ciphers: Optional[str] = None, username: Optional[str] = None,
                 password: Optional[str] = None, listeners=None, *args, **kwargs):
        """
        :param host: MQTT broker host
        :param port: MQTT broker port (default: 1883)
        :param topic: Topic to read messages from (default: ``platypush_bus_mq/<device_id>``)
        :param subscribe_default_topic: Whether the backend should subscribe the default topic (default:
            ``platypush_bus_mq/<device_id>``) and execute the messages received there as action requests
            (default: True).
        :param tls_cafile: If TLS/SSL is enabled on the MQTT server and the certificate requires a certificate authority
            to authenticate it, `ssl_cafile` will point to the provided ca.crt file (default: None)
        :param tls_certfile: If TLS/SSL is enabled on the MQTT server and a client certificate it required, specify it
            here (default: None)
        :param tls_keyfile: If TLS/SSL is enabled on the MQTT server and a client certificate key it required,
            specify it here (default: None) :type tls_keyfile: str
        :param tls_version: If TLS/SSL is enabled on the MQTT server and it requires a certain TLS version, specify it
            here (default: None)
        :param tls_ciphers: If TLS/SSL is enabled on the MQTT server and an explicit list of supported ciphers is
            required, specify it here (default: None)
        :param username: Specify it if the MQTT server requires authentication (default: None)
        :param password: Specify it if the MQTT server requires authentication (default: None)
        :param listeners: If specified then the MQTT backend will also listen for
            messages on the additional configured message queues. This parameter
            is a list of maps where each item supports the same arguments passed
            to the main backend configuration (host, port, topic, password etc.).
            Note that the message queue configured on the main configuration
            will expect valid Platypush messages that then can execute, while
            message queues registered to the listeners will accept any message. Example::

                listeners:
                    - host: localhost
                      topics:
                          - topic1
                          - topic2
                          - topic3
                    - host: sensors
                      topics:
                          - topic4
                          - topic5

        """

        super().__init__(*args, **kwargs)

        self.host = host
        self.port = port
        self.topic = '{}/{}'.format(topic, self.device_id)
        self.subscribe_default_topic = subscribe_default_topic
        self.username = username
        self.password = password
        self._client = None
        self._listeners = []

        self.tls_cafile = os.path.abspath(os.path.expanduser(tls_cafile)) \
            if tls_cafile else None

        self.tls_certfile = os.path.abspath(os.path.expanduser(tls_certfile)) \
            if tls_certfile else None

        self.tls_keyfile = os.path.abspath(os.path.expanduser(tls_keyfile)) \
            if tls_keyfile else None

        self.tls_version = tls_version
        self.tls_ciphers = tls_ciphers
        self.listeners_conf = listeners or []

    def send_message(self, msg, topic: Optional[str] = None, **kwargs):
        try:
            client = get_plugin('mqtt')
            client.send_message(topic=topic or self.topic, msg=msg, host=self.host,
                                port=self.port, username=self.username,
                                password=self.password, tls_cafile=self.tls_cafile,
                                tls_certfile=self.tls_certfile,
                                tls_keyfile=self.tls_keyfile,
                                tls_version=self.tls_version,
                                tls_ciphers=self.tls_ciphers, **kwargs)
        except Exception as e:
            self.logger.exception(e)

    @staticmethod
    def on_connect(*topics):
        # noinspection PyUnusedLocal
        def handler(client, userdata, flags, rc):
            for topic in topics:
                client.subscribe(topic)

        return handler

    def on_mqtt_message(self):
        def handler(client, _, msg):
            data = msg.payload
            # noinspection PyBroadException
            try:
                data = data.decode('utf-8')
                data = json.loads(data)
            except:
                pass

            # noinspection PyProtectedMember
            self.bus.post(MQTTMessageEvent(host=client._host, port=client._port, topic=msg.topic, msg=data))

        return handler

    def _initialize_listeners(self, listeners_conf):
        import paho.mqtt.client as mqtt

        # noinspection PyShadowingNames
        def listener_thread(client, host, port):
            client.connect(host, port)
            client.loop_forever()

        # noinspection PyShadowingNames,PyUnusedLocal
        for i, listener in enumerate(listeners_conf):
            host = listener.get('host')
            port = listener.get('port', self._default_mqtt_port)
            topics = listener.get('topics')
            username = listener.get('username')
            password = listener.get('password')
            tls_cafile = listener.get('tls_cafile')

            if not host or not topics:
                self.logger.warning('No host nor list of topics specified for ' +
                                    'listener n.{}'.format(i + 1))
                continue

            client = mqtt.Client()
            client.on_connect = self.on_connect(*topics)
            client.on_message = self.on_mqtt_message()

            if username and password:
                client.username_pw_set(username, password)

            if tls_cafile:
                client.tls_set(ca_certs=tls_cafile,
                               certfile=listener.get('tls_certfile'),
                               keyfile=listener.get('tls_keyfile'),
                               tls_version=listener.get('tls_version'),
                               ciphers=listener.get('tls_ciphers'))

            threading.Thread(target=listener_thread, kwargs={
                'client': client, 'host': host, 'port': port}).start()

    def on_exec_message(self):
        def handler(_, __, msg):
            # noinspection PyShadowingNames
            def response_thread(msg):
                set_thread_name('MQTTProcessor')
                response = self.get_message_response(msg)
                if not response:
                    return
                response_topic = '{}/responses/{}'.format(self.topic, msg.id)

                self.logger.info('Processing response on the MQTT topic {}: {}'.
                                 format(response_topic, response))

                self.send_message(response)

            msg = msg.payload.decode('utf-8')
            # noinspection PyBroadException
            try:
                msg = json.loads(msg)
                msg = Message.build(msg)
            except:
                pass

            if not msg:
                return

            self.logger.info('Received message on the MQTT backend: {}'.format(msg))

            try:
                self.on_message(msg)
            except Exception as e:
                self.logger.exception(e)
                return

            if isinstance(msg, Request):
                threading.Thread(target=response_thread, name='MQTTProcessor', args=(msg,)).start()

        return handler

    def run(self):
        import paho.mqtt.client as mqtt

        super().run()
        self._client = None

        if self.host:
            self._client = mqtt.Client()
            if self.subscribe_default_topic:
                self._client.on_connect = self.on_connect(self.topic)

            self._client.on_message = self.on_exec_message()
            if self.username and self.password:
                self._client.username_pw_set(self.username, self.password)

            if self.tls_cafile:
                self._client.tls_set(ca_certs=self.tls_cafile, certfile=self.tls_certfile,
                                     keyfile=self.tls_keyfile, tls_version=self.tls_version,
                                     ciphers=self.tls_ciphers)

            self._client.connect(self.host, self.port, 60)
            self.logger.info('Initialized MQTT backend on host {}:{}, topic {}'.
                             format(self.host, self.port, self.topic))

        self._initialize_listeners(self.listeners_conf)
        if self._client:
            self._client.loop_forever()

    def stop(self):
        self.logger.info('Received STOP event on MqttBackend')
        if self._client:
            self._client.disconnect()
            self._client.loop_stop()
            self._client = None

        for listener in self._listeners:
            try:
                listener.loop_stop()
            except Exception as e:
                # noinspection PyProtectedMember
                self.logger.warning('Could not stop listener {host}:{port}: {error}'.format(
                    host=listener._host, port=listener._port,
                    error=str(e)))


# vim:sw=4:ts=4:et:



import importlib
import time

from threading import Thread

from platypush.bus import Bus
from platypush.backend import Backend
from platypush.backend.http.request import HttpRequest
from platypush.message.request import Request


class HttpPollBackend(Backend):
    """
    This backend will poll multiple HTTP endpoints/services and return events
    the bus whenever something new happened. Supported types:
    :class:`platypush.backend.http.request.JsonHttpRequest` (for polling updates on
    a JSON endpoint), :class:`platypush.backend.http.request.rss.RssUpdates`
    (for polling updates on an RSS feed). Example configuration::

        backend.http.poll:
            requests:
                -
                    # Poll for updates on a JSON endpoint
                    method: GET
                    type: platypush.backend.http.request.JsonHttpRequest
                    args:
                        url: https://host.com/api/v1/endpoint
                        headers:
                            Token: TOKEN
                        params:
                            updatedSince: 1m
                        timeout: 5  # Times out after 5 seconds (default)
                    poll_seconds: 60  # Check for updates on this endpoint every 60 seconds (default)
                    path: ${response['items']}  # Path in the JSON to check for new items.
                                                # Python expressions are supported.
                                                # Note that 'response' identifies the JSON root.
                                                # Default value: JSON root.
                -
                    # Poll for updates on an RSS feed
                    type: platypush.backend.http.request.rss.RssUpdates
                    url: http://www.theguardian.com/rss/world
                    title: The Guardian - World News
                    poll_seconds: 120
                    max_entries: 10

    Triggers: an update event for the relevant HTTP source if it contains new items. For example:

        * :class:`platypush.message.event.http.rss.NewFeedEvent` if a feed contains new items
        * :class:`platypush.message.event.http.HttpEvent` if a JSON endpoint contains new items
    """

    def __init__(self, requests, *args, **kwargs):
        """
        :param requests: Configuration of the requests to make (see class description for examples)
        :type requests: dict
        """

        super().__init__(*args, **kwargs)
        self.requests = []

        for request in requests:
            if isinstance(request, dict):
                type = request['type']
                (module, name) = ('.'.join(type.split('.')[:-1]), type.split('.')[-1])
                module = importlib.import_module(module)
                request = getattr(module, name)(**request)
            elif not isinstance(request, HttpRequest):
                raise RuntimeError('Request should either be a dict or a ' +
                                   'HttpRequest object, {} found'.format(type(request)))

            request.bus = self.bus
            self.requests.append(request)


    def run(self):
        super().run()

        while not self.should_stop():
            for request in self.requests:
                if time.time() - request.last_request_timestamp > request.poll_seconds:
                    request.execute()

            time.sleep(0.1)  # Prevent a tight loop


# vim:sw=4:ts=4:et:




import functools
import mimetypes
import os

from platypush.utils import get_mime_type

from . import MediaHandler


class FileHandler(MediaHandler):
    prefix_handlers = ['file://']

    def __init__(self, source, *args, **kwargs):
        super().__init__(source, *args, **kwargs)

        self.path = os.path.abspath(os.path.expanduser(
            self.source[len(self._matched_handler):]))
        self.filename = self.path.split('/')[-1]

        if not os.path.isfile(self.path):
            raise FileNotFoundError('{} is not a valid file'.
                                    format(self.path))

        self.mime_type = get_mime_type(source)
        if self.mime_type[:5] not in ['video', 'audio', 'image']:
            raise AttributeError('{} is not a valid media file (detected format: {})'.
                                 format(source, self.mime_type))

        self.extension = mimetypes.guess_extension(self.mime_type)
        if self.url:
            self.url += self.extension
        self.content_length = os.path.getsize(self.path)


    def get_data(self, from_bytes=None, to_bytes=None, chunk_size=None):
        if from_bytes is None:
            from_bytes = 0
        if to_bytes is None:
            to_bytes = os.path.getsize(self.path)
        if chunk_size is None:
            chunk_size = os.path.getsize(self.path) - from_bytes

        with open(self.path, 'rb') as f:
            f.seek(from_bytes)
            for chunk in iter(functools.partial(
                    f.read, min(to_bytes-from_bytes, chunk_size)), b''):
                yield chunk


# vim:sw=4:ts=4:et:



import os
from typing import Optional

import pyotp

from platypush import Response
from platypush.config import Config
from platypush.plugins import Plugin, action


class OtpPlugin(Plugin):
    """
    This plugin can be used to generate OTP (One-Time Password) codes compatible with Google Authenticator and
    other 2FA (Two-Factor Authentication) applications.

    Requires:

        * **pyotp** (``pip install pyotp``)
    """

    def __init__(self, secret: Optional[str] = None, secret_path: Optional[str] = None,
                 provisioning_name: Optional[str] = None, issuer_name: Optional[str] = None, **kwargs):
        """
        :param secret: Base32-encoded secret to be used for password generation.
        :param secret_path: If no secret is provided statically, then it will be read from this path
            (default: ``~/.local/share/platypush/otp/secret``). If no secret is found then one will be
            generated.
        :param provisioning_name: If you want to use the Google Authenticator, you can specify the default
            email address to associate to your OTPs for the provisioning process here.
        :param issuer_name: If you want to use the Google Authenticator, you can specify the default
            issuer name to display on your OTPs here.
        """
        super().__init__(**kwargs)
        if not secret_path:
            secret_path = os.path.join(Config.get('workdir'), 'otp', 'secret')

        self.secret_path = secret_path
        self.secret = secret
        self.provisioning_name = provisioning_name
        self.issuer_name = issuer_name

    def _get_secret_from_path(self, secret_path: str) -> str:
        if not os.path.isfile(secret_path):
            secret = self.refresh_secret(secret_path).output
        else:
            with open(secret_path, 'r') as f:
                secret = f.readline()

        return secret

    def _get_secret(self, secret: Optional[str] = None, secret_path: Optional[str] = None) -> str:
        if secret:
            return secret
        if secret_path:
            return self._get_secret_from_path(secret_path)
        if self.secret:
            return self.secret
        if self.secret_path:
            return self._get_secret_from_path(self.secret_path)

        raise AssertionError('No secret nor secret_file specified')

    def _get_topt(self, secret: Optional[str] = None, secret_path: Optional[str] = None) -> pyotp.TOTP:
        return pyotp.TOTP(self._get_secret(secret, secret_path))

    def _get_hopt(self, secret: Optional[str] = None, secret_path: Optional[str] = None) -> pyotp.HOTP:
        return pyotp.HOTP(self._get_secret(secret, secret_path))

    @action
    def refresh_secret(self, secret_path: Optional[str] = None) -> Response:
        """
        Refresh the secret token for key generation given a secret path.

        :param secret_path: Secret path to refresh (default: default configured path).
        """

        secret_path = secret_path or self.secret_path
        assert secret_path, 'No secret_path configured'

        os.makedirs(os.path.dirname(os.path.abspath(os.path.expanduser(secret_path))), exist_ok=True)
        secret = pyotp.random_base32()
        with open(secret_path, 'w') as f:
            f.writelines([secret])
        os.chmod(secret_path, 0o600)
        return secret

    @action
    def get_time_otp(self, secret: Optional[str] = None, secret_path: Optional[str] = None) -> str:
        """
        :param secret: Secret token to be used (overrides configured ``secret``).
        :param secret_path: File containing the secret to be used (overrides configured ``secret_path``).
        :return: A time-based token, as a string.
        """
        otp = self._get_topt(secret, secret_path)
        return otp.now()

    @action
    def get_counter_otp(self, count: int, secret: Optional[str] = None, secret_path: Optional[str] = None) -> str:
        """
        :param count: Index for the counter-OTP.
        :param secret: Secret token to be used (overrides configured ``secret``).
        :param secret_path: File containing the secret to be used (overrides configured ``secret_path``).
        :return: A count-based token, as a string.
        """
        otp = self._get_hopt(secret, secret_path)
        return otp.at(count)

    @action
    def verify_time_otp(self, otp: str, secret: Optional[str] = None, secret_path: Optional[str] = None) -> bool:
        """
        Verify a code against a stored time-OTP.

        :param otp: Code to be verified.
        :param secret: Secret token to be used (overrides configured ``secret``).
        :param secret_path: File containing the secret to be used (overrides configured ``secret_path``).
        :return: True if the code is valid, False otherwise.
        """
        _otp = self._get_topt(secret, secret_path)
        return _otp.verify(otp)

    @action
    def verify_counter_otp(self, otp: str, count: int, secret: Optional[str] = None,
                           secret_path: Optional[str] = None) -> bool:
        """
        Verify a code against a stored counter-OTP.

        :param otp: Code to be verified.
        :param count: Index for the counter-OTP to be verified.
        :param secret: Secret token to be used (overrides configured ``secret``).
        :param secret_path: File containing the secret to be used (overrides configured ``secret_path``).
        :return: True if the code is valid, False otherwise.
        """
        _otp = self._get_hopt(secret, secret_path)
        return _otp.verify(otp, count)

    @action
    def provision_time_otp(self, name: Optional[str] = None, issuer_name: Optional[str] = None,
                           secret: Optional[str] = None, secret_path: Optional[str] = None) -> str:
        """
        Generate a provisioning URI for a time-OTP that can be imported in Google Authenticator.

        :param name: Name or e-mail address associated to the account used by the Google Authenticator.
            If None is specified then the value will be read from the configured ``provisioning_name``.
        :param issuer_name: Name of the issuer of the OTP (default: default configured ``issuer_name`` or None).
        :param secret: Secret token to be used (overrides configured ``secret``).
        :param secret_path: File containing the secret to be used (overrides configured ``secret_path``).
        :return: Generated provisioning URI.
        """
        name = name or self.provisioning_name
        issuer_name = issuer_name or self.issuer_name
        assert name, 'No account name or default provisioning address provided'

        _otp = self._get_topt(secret, secret_path)
        return _otp.provisioning_uri(name, issuer_name=issuer_name)

    @action
    def provision_counter_otp(self, name: Optional[str] = None, issuer_name: Optional[str] = None, initial_count=0,
                              secret: Optional[str] = None, secret_path: Optional[str] = None) -> str:
        """
        Generate a provisioning URI for a counter-OTP that can be imported in Google Authenticator.

        :param name: Name or e-mail address associated to the account used by the Google Authenticator.
            If None is specified then the value will be read from the configured ``provisioning_name``.
        :param issuer_name: Name of the issuer of the OTP (default: default configured ``issuer_name`` or None).
        :param initial_count: Initial value for the counter (default: 0).
        :param secret: Secret token to be used (overrides configured ``secret``).
        :param secret_path: File containing the secret to be used (overrides configured ``secret_path``).
        :return: Generated provisioning URI.
        """
        name = name or self.provisioning_name
        issuer_name = issuer_name or self.issuer_name
        assert name, 'No account name or default provisioning address provided'

        _otp = self._get_hopt(secret, secret_path)
        return _otp.provisioning_uri(name, issuer_name=issuer_name, initial_count=initial_count)


# vim:sw=4:ts=4:et:



import time

from threading import Timer

from platypush.backend import Backend
from platypush.message.event.midi import MidiMessageEvent


class MidiBackend(Backend):
    """
    This backend will listen for events from a MIDI device and post a
    MidiMessageEvent whenever a new MIDI event happens.

    Triggers:

        * :class:`platypush.message.event.midi.MidiMessageEvent` when a new MIDI event is received

    Requires:

        * **rtmidi** (``pip install rtmidi``)
    """

    def __init__(self, device_name=None, port_number=None,
                 midi_throttle_time=None, *args, **kwargs):
        """
        :param device_name: Name of the MIDI device.  *N.B.* either
            `device_name` or `port_number` must be set.
            Use :meth:`platypush.plugins.midi.query_ports` to get the
            available ports indices and names
        :type device_name: str

        :param port_number: MIDI port number
        :type port_number: int

        :param midi_throttle_time: If set, the MIDI events will be throttled -
            max one per selected time frame (in seconds). Set this parameter if
            you want to synchronize MIDI events with plugins that normally
            operate with a lower throughput.
        :type midi_throttle_time: int
        """

        import rtmidi
        super().__init__(*args, **kwargs)

        if (device_name and port_number is not None) or \
                (not device_name and port_number is None):
            raise RuntimeError('Either device_name or port_number (not both) ' +
                               'must be set in the MIDI backend configuration')

        self.midi_throttle_time = midi_throttle_time
        self.midi = rtmidi.MidiIn()
        self.last_trigger_event_time = None
        self.midi_flush_timeout = None
        ports = self.midi.get_ports()

        if not ports:
            raise RuntimeError('No MIDI devices available')

        if device_name:
            if device_name not in ports:
                raise RuntimeError('MIDI device "{}" not found'.format(device_name))

            self.port_number = ports.index(device_name)
            self.device_name = device_name

        if port_number:
            if port_number < 0 or port_number >= len(ports):
                raise RuntimeError('MIDI port {} not found')

            self.port_number = port_number
            self.device_name = ports[port_number]

        self.midi.set_callback(self._on_midi_message())


    def _on_midi_message(self):
        def flush_midi_message(message):
            def _f():
                self.logger.info('Flushing throttled MIDI message {} to the bus'.format(message))
                delay = time.time() - self.last_trigger_event_time
                self.bus.post(MidiMessageEvent(message=message, delay=delay))
            return _f

        def callback(message, data):
            # rtmidi will provide a tuple in the format
            # (midi_message, time_since_last_event)
            delay = message[1]
            message = message[0]

            if self.midi_throttle_time and self.last_trigger_event_time:
                event_delta = time.time() - self.last_trigger_event_time
                if event_delta < self.midi_throttle_time:
                    self.logger.debug('Skipping throttled message {}'.format(message))
                    if self.midi_flush_timeout:
                        self.midi_flush_timeout.cancel()

                    self.midi_flush_timeout = Timer(
                        self.midi_throttle_time-event_delta,
                        flush_midi_message(message))

                    self.midi_flush_timeout.start()
                    return

            self.last_trigger_event_time = time.time()
            self.bus.post(MidiMessageEvent(message=message, delay=delay))

        return callback


    def run(self):
        super().run()

        self.midi.open_port(self.port_number)
        self.logger.info('Initialized MIDI backend, listening for events on device {}'.
                     format(self.device_name))

        while not self.should_stop():
            try:
                time.sleep(1)
            except Exception as e:
                self.logger.exception(e)

        if self.midi:
            self.midi.close_port(self.port_number)
            self.midi = None


# vim:sw=4:ts=4:et:




import time

from platypush.plugins import Plugin, action

class LastfmPlugin(Plugin):
    """
    Plugin to interact with your Last.FM (https://last.fm) account, update your
    current track and your scrobbles.

    Requires:

        * **pylast** (``pip install pylast``)
    """

    def __init__(self, api_key, api_secret, username, password):
        """
        :param api_key: Last.FM API key, see https://www.last.fm/api
        :type api_key: str

        :param api_secret: Last.FM API secret, see https://www.last.fm/api
        :type api_key: str

        :param username: Last.FM username
        :type api_key: str

        :param password: Last.FM password, used to sign the requests
        :type api_key: str
        """

        import pylast
        super().__init__()

        self.api_key = api_key
        self.api_secret = api_secret
        self.username = username
        self.password = password

        self.lastfm = pylast.LastFMNetwork(
            api_key = self.api_key,
            api_secret = self.api_secret,
            username = self.username,
            password_hash = pylast.md5(self.password))


    @action
    def scrobble(self, artist, title, album=None, **kwargs):
        """
        Scrobble a track to Last.FM

        :param artist: Artist
        :type artist: str
        :param title: Title
        :type title: str
        :param album: Album (optional)
        :type album: str
        """

        self.lastfm.scrobble(
            artist = artist,
            title = title,
            album = album,
            timestamp = int(time.time()),
        )


    @action
    def update_now_playing(self, artist, title, album=None, **kwargs):
        """
        Update the currently playing track

        :param artist: Artist
        :type artist: str
        :param title: Title
        :type title: str
        :param album: Album (optional)
        :type album: str
        """

        self.lastfm.update_now_playing(
            artist = artist,
            title = title,
            album = album,
        )


# vim:sw=4:ts=4:et:




import argparse
import logging
import re
import sys

from platypush.bus import Bus
from platypush.config import Config
from platypush.context import register_backends
from platypush.message.request import Request


class Pusher(object):
    """
    Main class to send messages and events to a node
    """

    """ Configuration file path """
    config_file = None

    """ Default backend name """
    backend = None

    """ Pusher local bus. The response will be processed here """
    bus = None

    """ Configured backends as a name => object map """
    backends = {}

    """ Default response_wait timeout """
    default_response_wait_timeout = 5


    def __init__(self, config_file=None, backend=None, on_response=None):
        """
        Constructor.
        Params:
            config_file -- Path to the configuration file - default:
                           ~/.config/platypush/config.yaml or
                           /etc/platypush/config.yaml)
            backend     -- Name of the backend where pusher will send the
                           request and wait for the response (kafka
                           or pushbullet). Default: whatever is specified
                           with pusher=true in your configuration file
            on_response -- Method that will be invoked upon response receipt.
                           Takes a platypush.message.response.Response as arg.
                           Default: print the response and exit.
        """

        # Initialize the configuration
        self.config_file = config_file
        log_conf = Config.get('logging')
        Config.init(config_file)
        logging.basicConfig(level=log_conf['level']
                            if log_conf and 'level' in log_conf
                            else logging.info, stream=sys.stdout)

        self.on_response = on_response or self.default_on_response()
        self.backend = backend or Config.get_default_pusher_backend()
        self.bus = Bus()


    @classmethod
    def parse_build_args(cls, args):
        """ Parse the recognized options from a list of cmdline arguments """
        parser = argparse.ArgumentParser()
        parser.add_argument('--config', '-c', dest='config', required=False,
                            default=None, help="Configuration file path (default: " +
                            "~/.config/platypush/config.yaml or " +
                            "/etc/platypush/config.yaml")

        parser.add_argument('--type', '-p', dest='type', required=False,
                            default='request', help="Type of message to send, request or event")

        parser.add_argument('--target', '-t', dest='target', required=False,
                            default=Config.get('device_id'),
                            help="Destination of the command")

        parser.add_argument('--action', '-a', dest='action', required=False,
                            default=None, help="Action to execute, as " +
                            "package.method (e.g. music.mpd.play), if this is a request")

        parser.add_argument('--event', '-e', dest='event', required=False,
                            default=None, help="Event type, as full " +
                            "package.class (e.g. " +
                            "platypush.message.event.ping.PingEvent), if this is an event")

        parser.add_argument('--backend', '-b', dest='backend', required=False,
                            default=None, help="Backend to deliver the message " +
                            "[pushbullet|kafka] (default: whatever " +
                            "specified in your config with pusher=True)")

        parser.add_argument('--timeout', '-T', dest='timeout', required=False,
                            default=cls.default_response_wait_timeout, help="The application " +
                            "will wait for a response for this number of seconds " +
                            "(default: " + str(cls.default_response_wait_timeout) + " seconds. "
                            "A zero value means that the application " +
                            " will exit without waiting for a response)")

        opts, args = parser.parse_known_args(args)

        if len(args) % 2 != 0:
            parser.print_help()
            raise RuntimeError('Odd number of key-value options passed: {}'.format(args))

        if opts.type == 'request' and not opts.action:
            parser.print_help()
            raise RuntimeError('No action provided for the request'.format(args))

        if opts.type == 'event' and not opts.event:
            parser.print_help()
            raise RuntimeError('No type provided for the event'.format(args))

        opts.args = {}
        for i in range(0, len(args), 2):
            opts.args[re.sub('^-+', '', args[i])] = args[i+1]

        return opts

    def get_backend(self, name):
        # Lazy init
        if not self.backends:
            self.backends = register_backends(bus=self.bus)

        if name not in self.backends:
            raise RuntimeError('No such backend configured: {}'.format(name))
        return self.backends[name]

    def default_on_response(self):
        def _f(response):
            logging.info('Received response: {}'.format(response))
            # self.backend_instance.stop()
        return _f

    def send_event(self, target=Config.get('device_id'),
                   type='platypush.message.event.Event', backend=None, **kwargs):
        if not backend: backend = self.backend

        self.backend_instance = self.get_backend(backend)
        self.backend_instance.send_event({
            'target': target,
            'args': {
                'type': type,
                **kwargs
            }
        })


    def send_request(self, target, action, backend=None,
            timeout=default_response_wait_timeout, **kwargs):
        """
        Sends a message on a backend and optionally waits for an answer.
        Params:
            target  -- Target node
            action  -- Action to be executed in the form plugin.path.method
                    (e.g. shell.exec or music.mpd.play)
            backend -- Name of the backend that will process the request and get
                    the response (e.g. 'pushbullet' or 'kafka') (default: whichever
                    backend marked as pusher=true in your config.yaml)
            timeout -- Response receive timeout in seconds
                    - Pusher Default: 5 seconds
                    - If timeout == 0 or None: Pusher exits without waiting for a response
            **kwargs    -- Optional key-valued arguments for the action method
                        (e.g. cmd='echo ping' or groups="['Living Room']")
        """

        def _timeout_hndl(signum, frame):
            """ Default response timeout handle: raise RuntimeError and exit """

        if not backend: backend = self.backend

        req = Request.build({
            'target' : target,
            'action' : action,
            'args'   : kwargs,
        })

        self.backend_instance = self.get_backend(backend)
        self.backend_instance.send_request(req, on_response=self.on_response,
                                           response_timeout=timeout)


def main():
    opts = Pusher.parse_build_args(sys.argv[1:])
    pusher = Pusher(config_file=opts.config, backend=opts.backend)

    if opts.type == 'event':
        pusher.send_event(target=opts.target, type=opts.event, **opts.args)
    else:
        pusher.send_request(target=opts.target, action=opts.action, timeout=opts.timeout, **opts.args)


# vim:sw=4:ts=4:et:




import logging

class MediaHandler:
    """
    Abstract class to manage media handlers that can be streamed over the HTTP
    server through the `/media` endpoint.
    """

    prefix_handlers = []

    def __init__(self, source, filename=None,
                 mime_type='application/octet-stream', name=None, url=None,
                 subtitles=None):
        matched_handlers = [hndl for hndl in self.prefix_handlers
                            if source.startswith(hndl)]

        if not matched_handlers:
            raise AttributeError(('No matched handlers found for source "{}" ' +
                                  'through {}. Supported handlers: {}').format(
                                      source, self.__class__.__name__,
                                      self.prefix_handlers))

        self.name = name
        self.path = None
        self.filename = filename
        self.source = source
        self.url = url
        self.mime_type = mime_type
        self.subtitles = subtitles
        self.content_length = 0
        self._matched_handler = matched_handlers[0]

    @classmethod
    def build(cls, source, *args, **kwargs):
        errors = {}

        for hndl_class in supported_handlers:
            try:
                return hndl_class(source, *args, **kwargs)
            except Exception as e:
                logging.exception(e)
                errors[hndl_class.__name__] = str(e)

        raise AttributeError(('The source {} has no handlers associated. ' +
                              'Errors: {}').format(source, errors))

    def get_data(self, from_bytes=None, to_bytes=None, chunk_size=None):
        raise NotImplementedError()

    def set_subtitles(self, subtitles_file):
        self.subtitles = subtitles_file

    def remove_subtitles(self):
        self.subtitles = None

    def __iter__(self):
        for attr in ['name', 'source', 'mime_type', 'url', 'subtitles',
                     'prefix_handlers', 'media_id']:
            if hasattr(self, attr):
                yield (attr, getattr(self, attr))


from .file import FileHandler

__all__ = ['MediaHandler', 'FileHandler']


supported_handlers = [eval(hndl) for hndl in __all__
                      if hndl != MediaHandler.__name__]


# vim:sw=4:ts=4:et:



import datetime

from flask import Blueprint, request, redirect, render_template, make_response, url_for

from platypush.backend.http.app import template_folder
from platypush.backend.http.utils import HttpUtils
from platypush.user import UserManager

register = Blueprint('register', __name__, template_folder=template_folder)

# Declare routes list
__routes__ = [
    register,
]


@register.route('/register', methods=['GET', 'POST'])
def register():
    """ Registration page """
    user_manager = UserManager()
    redirect_page = request.args.get('redirect', '/')
    session_token = request.cookies.get('session_token')

    if session_token:
        user, session = user_manager.authenticate_user_session(session_token)
        if user:
            return redirect(redirect_page, 302)

    if request.form:
        username = request.form.get('username')
        password = request.form.get('password')
        confirm_password = request.form.get('confirm_password')
        remember = request.form.get('remember')

        if password == confirm_password:
            user_manager.create_user(username=username, password=password)
            session = user_manager.create_user_session(username=username, password=password,
                                                       expires_at=datetime.datetime.utcnow() + datetime.timedelta(days=1)
                                                       if not remember else None)

            if session:
                redirect_target = redirect(redirect_page, 302)
                response = make_response(redirect_target)
                response.set_cookie('session_token', session.session_token)
                return response

    if user_manager.get_user_count() > 0:
        return redirect('/login?redirect=' + redirect_page, 302)

    return render_template('register.html', utils=HttpUtils)


# vim:sw=4:ts=4:et:



import datetime

from typing import Optional, List

from platypush.message.response import Response


class TelegramMessageResponse(Response):
    def __init__(self,
                 message_id: int,
                 chat_id: int,
                 creation_date: Optional[datetime.datetime],
                 chat_username: Optional[str] = None,
                 chat_firstname: Optional[str] = None,
                 chat_lastname: Optional[str] = None,
                 from_user_id: Optional[int] = None,
                 from_username: Optional[str] = None,
                 from_firstname: Optional[str] = None,
                 from_lastname: Optional[str] = None,
                 text: Optional[str] = None,
                 caption: Optional[str] = None,
                 edit_date: Optional[datetime.datetime] = None,
                 forward_date: Optional[datetime.datetime] = None,
                 forward_from_message_id: Optional[int] = None,
                 photo_file_id: Optional[str] = None,
                 photo_file_size: Optional[int] = None,
                 photo_width: Optional[int] = None,
                 photo_height: Optional[int] = None,
                 document_file_id: Optional[str] = None,
                 document_file_name: Optional[str] = None,
                 document_file_size: Optional[str] = None,
                 document_mime_type: Optional[str] = None,
                 audio_file_id: Optional[str] = None,
                 audio_file_size: Optional[str] = None,
                 audio_mime_type: Optional[str] = None,
                 audio_performer: Optional[str] = None,
                 audio_title: Optional[str] = None,
                 audio_duration: Optional[str] = None,
                 location_latitude: Optional[float] = None,
                 location_longitude: Optional[float] = None,
                 contact_phone_number: Optional[str] = None,
                 contact_first_name: Optional[str] = None,
                 contact_last_name: Optional[str] = None,
                 contact_user_id: Optional[int] = None,
                 contact_vcard: Optional[str] = None,
                 video_file_id: Optional[str] = None,
                 video_file_size: Optional[int] = None,
                 video_width: Optional[int] = None,
                 video_height: Optional[int] = None,
                 video_mime_type: Optional[str] = None,
                 video_duration: Optional[str] = None,
                 link: Optional[str] = None,
                 media_group_id: Optional[int] = None,
                 *args, **kwargs):
        super().__init__(*args, output={
            'message_id': message_id,
            'chat_id': chat_id,
            'chat_username': chat_username,
            'chat_firstname': chat_firstname,
            'chat_lastname': chat_lastname,
            'from_user_id': from_user_id,
            'from_username': from_username,
            'from_firstname': from_firstname,
            'from_lastname': from_lastname,
            'text': text,
            'caption': caption,
            'creation_date': creation_date,
            'edit_date': edit_date,
            'forward_from_message_id': forward_from_message_id,
            'forward_date': forward_date,
            'photo_file_id': photo_file_id,
            'photo_file_size': photo_file_size,
            'photo_width': photo_width,
            'photo_height': photo_height,
            'document_file_id': document_file_id,
            'document_file_name': document_file_name,
            'document_file_size': document_file_size,
            'document_mime_type': document_mime_type,
            'audio_file_id': audio_file_id,
            'audio_file_size': audio_file_size,
            'audio_performer': audio_performer,
            'audio_title': audio_title,
            'audio_duration': audio_duration,
            'audio_mime_type': audio_mime_type,
            'video_file_id': video_file_id,
            'video_file_size': video_file_size,
            'video_width': video_width,
            'video_height': video_height,
            'video_duration': video_duration,
            'video_mime_type': video_mime_type,
            'link': link,
            'location_latitude': location_latitude,
            'location_longitude': location_longitude,
            'contact_phone_number': contact_phone_number,
            'contact_first_name': contact_first_name,
            'contact_last_name': contact_last_name,
            'contact_user_id': contact_user_id,
            'contact_vcard': contact_vcard,
            'media_group_id': media_group_id,
        }, **kwargs)


class TelegramFileResponse(Response):
    def __init__(self,
                 file_id: str,
                 file_path: str,
                 file_size: int,
                 *args, **kwargs):
        super().__init__(*args, output={
            'file_id': file_id,
            'file_path': file_path,
            'file_size': file_size,
        }, **kwargs)


class TelegramChatResponse(Response):
    # noinspection PyShadowingBuiltins
    def __init__(self,
                 chat_id: int,
                 link: str,
                 username: str,
                 invite_link: Optional[str],
                 title: Optional[str] = None,
                 description: Optional[str] = None,
                 type: Optional[str] = None,
                 first_name: Optional[str] = None,
                 last_name: Optional[str] = None,
                 *args, **kwargs):
        super().__init__(*args, output={
            'chat_id': chat_id,
            'link': link,
            'invite_link': invite_link,
            'username': username,
            'title': title,
            'description': description,
            'type': type,
            'first_name': first_name,
            'last_name': last_name,
        }, **kwargs)


class TelegramUserResponse(Response):
    # noinspection PyShadowingBuiltins
    def __init__(self,
                 user_id: int,
                 username: str,
                 is_bot: bool,
                 first_name: str,
                 last_name: Optional[str] = None,
                 language_code: Optional[str] = None,
                 link: Optional[str] = None,
                 *args, **kwargs):
        super().__init__(*args, output={
            'user_id': user_id,
            'username': username,
            'is_bot': is_bot,
            'link': link,
            'language_code': language_code,
            'first_name': first_name,
            'last_name': last_name,
        }, **kwargs)


class TelegramUsersResponse(Response):
    # noinspection PyShadowingBuiltins
    def __init__(self,
                 users: List[TelegramUserResponse],
                 *args, **kwargs):
        super().__init__(*args, output=[user.output for user in users], **kwargs)


# vim:sw=4:ts=4:et:



import json
import urllib3
import urllib.request
import urllib.parse

from platypush.plugins import Plugin, action
from platypush.plugins.media import PlayerState

class VideoTorrentcastPlugin(Plugin):
    def __init__(self, server='localhost', port=9090, *args, **kwargs):
        self.server = server
        self.port = port
        self.state = PlayerState.STOP.value

    @action
    def play(self, url):
        request = urllib.request.urlopen(
            'http://{}:{}/play/'.format(self.server, self.port),
            data=urllib.parse.urlencode({
                'url': url
            }).encode()
        )

        self.state = PlayerState.PLAY.value
        return request.read()

    @action
    def pause(self):
        http = urllib3.PoolManager()
        request = http.request('POST',
            'http://{}:{}/pause/'.format(self.server, self.port))

        self.state = PlayerState.PAUSE.value
        return request.read()

    @action
    def stop(self):
        http = urllib3.PoolManager()
        request = http.request('POST',
            'http://{}:{}/stop/'.format(self.server, self.port))

        self.state = PlayerState.STOP.value
        return request.read()

    @action
    def search(self, query):
        request = urllib.request.urlopen(urllib.request.Request(
            'https://api.apidomain.info/list?' + urllib.parse.urlencode({
                'sort': 'relevance',
                'quality': '720p,1080p,3d',
                'page': 1,
                'keywords': query,
            }),
            headers = {
                'User-Agent': 'Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 ' +
                    '(KHTML, like Gecko) Chrome/62.0.3202.94 Safari/537.36'
            })
        )

        results = json.loads(request.read())
        return results

    @action
    def search_and_play(self, query):
        response = self.search(query)
        if not response.output['MovieList']:
            self.logger.info('No torrent results found for {}'.format(query))

        item = response.output['MovieList'][0]
        magnet = item['items'][0]['torrent_magnet']
        self.logger.info('Playing torrent "{}" from {}'
                     .format(item['title'], magnet))

        return self.play(magnet)

    @action
    def voldown(self): raise NotImplementedError()

    @action
    def volup(self): raise NotImplementedError()

    @action
    def back(self): raise NotImplementedError()

    @action
    def forward(self): raise NotImplementedError()

    @action
    def status(self): return { 'state': self.state }


# vim:sw=4:ts=4:et:




import json

from redis import Redis

from platypush.backend import Backend
from platypush.context import get_plugin
from platypush.message import Message


class RedisBackend(Backend):
    """
    Backend that reads messages from a configured Redis queue (default:
    ``platypush_bus_mq``) and posts them to the application bus.  Very
    useful when you have plugin whose code is executed in another process
    and can't post events or requests to the application bus.

    Requires:

        * **redis** (``pip install redis``)
    """

    def __init__(self, queue='platypush_bus_mq', redis_args=None, *args, **kwargs):
        """
        :param queue: Queue name to listen on (default: ``platypush_bus_mq``)
        :type queue: str

        :param redis_args: Arguments that will be passed to the redis-py constructor (e.g. host, port, password), see
            http://redis-py.readthedocs.io/en/latest/
        :type redis_args: dict
        """

        super().__init__(*args, **kwargs)
        if redis_args is None:
            redis_args = {}

        self.queue = queue

        if not redis_args:
            redis_plugin = get_plugin('redis')
            if redis_plugin and redis_plugin.kwargs:
                redis_args = redis_plugin.kwargs

        self.redis_args = redis_args
        self.redis = Redis(**self.redis_args)

    def send_message(self, msg, queue_name=None, **kwargs):
        msg = str(msg)
        if queue_name:
            self.redis.rpush(queue_name, msg)
        else:
            self.redis.rpush(self.queue, msg)

    # noinspection PyBroadException
    def get_message(self, queue_name=None):
        queue = queue_name or self.queue
        msg = self.redis.blpop(queue)[1].decode('utf-8')

        try:
            msg = Message.build(json.loads(msg))
        except:
            try:
                import ast
                msg = Message.build(ast.literal_eval(msg))
            except:
                try:
                    msg = json.loads(msg)
                except Exception as e:
                    self.logger.exception(e)

        return msg

    def run(self):
        super().run()

        self.logger.info('Initialized Redis backend on queue {} with arguments {}'.
                         format(self.queue, self.redis_args))

        while not self.should_stop():
            try:
                msg = self.get_message()
                self.logger.info('Received message on the Redis backend: {}'.
                                 format(msg))
                self.on_message(msg)
            except Exception as e:
                self.logger.exception(e)


# vim:sw=4:ts=4:et:



import inspect
import logging
import queue
import os
import threading

from typing import Optional

from platypush.backend import Backend
from platypush.config import Config
from platypush.message.event.zwave import ZwaveNetworkReadyEvent, ZwaveNetworkStoppedEvent, ZwaveEvent, \
    ZwaveNodeAddedEvent, ZwaveValueAddedEvent, ZwaveNodeQueryCompletedEvent, ZwaveValueChangedEvent, \
    ZwaveValueRefreshedEvent, ZwaveValueRemovedEvent, ZwaveNetworkResetEvent, ZwaveCommandEvent, \
    ZwaveCommandWaitingEvent, ZwaveNodeRemovedEvent, ZwaveNodeRenamedEvent, ZwaveNodeReadyEvent, \
    ZwaveButtonRemovedEvent, ZwaveButtonCreatedEvent, ZwaveButtonOnEvent, ZwaveButtonOffEvent, ZwaveNetworkErrorEvent, \
    ZwaveNodeGroupEvent, ZwaveNodePollingEnabledEvent, ZwaveNodePollingDisabledEvent, ZwaveNodeSceneEvent, \
    ZwaveNodeEvent

event_queue = queue.Queue()
network_ready = threading.Event()


class _ZWEvent:
    def __init__(self, signal: str, sender: str, network=None, **kwargs):
        self.signal = signal
        self.sender = sender
        self.network = network
        self.args = kwargs


def _zwcallback(signal, sender, network, **kwargs):
    if signal == network.SIGNAL_NETWORK_AWAKED:
        network_ready.set()
    event_queue.put(_ZWEvent(signal=signal, sender=sender, network=network, **kwargs))


class ZwaveBackend(Backend):
    """
    Start and manage a Z-Wave network.

    If you are using a USB adapter and want a consistent naming for the device paths, you can use udev.

    .. code-block:: shell

        # Get the vendorID and productID of your device through lsusb.
        # Then add a udev rule for it to link it e.g. to /dev/zwave.

        cat <<EOF > /etc/udev/rules.d/92-zwave.rules
        SUBSYSTEM=="tty", ATTRS{idVendor}=="0658", ATTRS{idProduct}=="0200", SYMLINK+="zwave"
        EOF

        # Restart the udev service
        systemctl restart systemd-udevd.service

    Triggers:

        * :class:`platypush.message.event.zwave.ZwaveNetworkReadyEvent` when the network is up and running.
        * :class:`platypush.message.event.zwave.ZwaveNetworkStoppedEvent` when the network goes down.
        * :class:`platypush.message.event.zwave.ZwaveNetworkResetEvent` when the network is reset.
        * :class:`platypush.message.event.zwave.ZwaveNetworkErrorEvent` when an error occurs on the network.
        * :class:`platypush.message.event.zwave.ZwaveNodeQueryCompletedEvent` when all the nodes on the network
          have been queried.
        * :class:`platypush.message.event.zwave.ZwaveNodeEvent` when a node attribute changes.
        * :class:`platypush.message.event.zwave.ZwaveNodeAddedEvent` when a node is added to the network.
        * :class:`platypush.message.event.zwave.ZwaveNodeRemovedEvent` when a node is removed from the network.
        * :class:`platypush.message.event.zwave.ZwaveNodeRenamedEvent` when a node is renamed.
        * :class:`platypush.message.event.zwave.ZwaveNodeReadyEvent` when a node is ready.
        * :class:`platypush.message.event.zwave.ZwaveNodeGroupEvent` when a node is associated/de-associated to a
          group.
        * :class:`platypush.message.event.zwave.ZwaveNodeSceneEvent` when a scene is set on a node.
        * :class:`platypush.message.event.zwave.ZwaveNodePollingEnabledEvent` when the polling is successfully turned
          on a node.
        * :class:`platypush.message.event.zwave.ZwaveNodePollingDisabledEvent` when the polling is successfully turned
          off a node.
        * :class:`platypush.message.event.zwave.ZwaveButtonCreatedEvent` when a button is added to the network.
        * :class:`platypush.message.event.zwave.ZwaveButtonRemovedEvent` when a button is removed from the network.
        * :class:`platypush.message.event.zwave.ZwaveButtonOnEvent` when a button is pressed.
        * :class:`platypush.message.event.zwave.ZwaveButtonOffEvent` when a button is released.
        * :class:`platypush.message.event.zwave.ZwaveValueAddedEvent` when a value is added to a node on the network.
        * :class:`platypush.message.event.zwave.ZwaveValueChangedEvent` when the value of a node on the network
          changes.
        * :class:`platypush.message.event.zwave.ZwaveValueRefreshedEvent` when the value of a node on the network
          is refreshed.
        * :class:`platypush.message.event.zwave.ZwaveValueRemovedEvent` when the value of a node on the network
          is removed.
        * :class:`platypush.message.event.zwave.ZwaveCommandEvent` when a command is received on the network.
        * :class:`platypush.message.event.zwave.ZwaveCommandWaitingEvent` when a command is waiting for a message
          to complete.

    Requires:

        * **python-openzwave** (``pip install python-openzwave``)

    """

    def __init__(self, device: str, config_path: Optional[str] = None, user_path: Optional[str] = None,
                 ready_timeout: float = 10.0, *args, **kwargs):
        """
        :param device: Path to the Z-Wave adapter (e.g. /dev/ttyUSB0 or /dev/ttyACM0).
        :param config_path: Z-Wave configuration path (default: ``<OPENZWAVE_PATH>/ozw_config``).
        :param user_path: Z-Wave user path where runtime and configuration files will be stored
            (default: ``<PLATYPUSH_WORKDIR>/zwave``).
        :param ready_timeout: Network ready timeout in seconds (default: 60).
        """
        import python_openzwave
        from openzwave.network import ZWaveNetwork

        super().__init__(*args, **kwargs)
        self.device = device

        if not config_path:
            config_path = os.path.join(os.path.dirname(inspect.getfile(python_openzwave)), 'ozw_config')
        if not user_path:
            user_path = os.path.join(Config.get('workdir'), 'zwave')
            os.makedirs(user_path, mode=0o770, exist_ok=True)

        self.config_path = config_path
        self.user_path = user_path
        self.ready_timeout = ready_timeout
        self.network: Optional[ZWaveNetwork] = None

    def start_network(self):
        if self.network and self.network.state >= self.network.STATE_AWAKED:
            self.logger.info('Z-Wave network already started')
            return

        from openzwave.network import ZWaveNetwork, dispatcher
        from openzwave.option import ZWaveOption

        network_ready.clear()
        logging.getLogger('openzwave').addHandler(self.logger)
        opts = ZWaveOption(self.device, config_path=self.config_path, user_path=self.user_path)
        opts.set_console_output(False)
        opts.lock()

        self.network = ZWaveNetwork(opts, log=None)
        dispatcher.connect(_zwcallback)
        ready = network_ready.wait(self.ready_timeout)

        if not ready:
            self.logger.warning('Driver not ready after {} seconds: continuing anyway'.format(self.ready_timeout))

    def stop_network(self):
        if self.network:
            self.network.stop()
            network_ready.clear()
            self.network = None

    def _process_event(self, event: _ZWEvent):
        from platypush.plugins.zwave import ZwavePlugin
        network = event.network or self.network

        if event.signal == network.SIGNAL_NETWORK_STOPPED or \
                event.signal == network.SIGNAL_DRIVER_REMOVED:
            event_queue.put(ZwaveNetworkStoppedEvent(device=self.device))
        elif event.signal == network.SIGNAL_ALL_NODES_QUERIED or \
                event.signal == network.SIGNAL_ALL_NODES_QUERIED_SOME_DEAD:
            event = ZwaveNodeQueryCompletedEvent(device=self.device)
        elif event.signal == network.SIGNAL_NETWORK_FAILED:
            event = ZwaveNetworkErrorEvent(device=self.device)
            self.logger.warning('Z-Wave network error')
        elif event.signal == network.SIGNAL_NETWORK_RESETTED or \
                event.signal == network.SIGNAL_DRIVER_RESET:
            event = ZwaveNetworkResetEvent(device=self.device)
        elif event.signal == network.SIGNAL_BUTTON_ON:
            event = ZwaveButtonOnEvent(device=self.device,
                                       node=ZwavePlugin.node_to_dict(event.args['node']))
        elif event.signal == network.SIGNAL_BUTTON_OFF:
            event = ZwaveButtonOffEvent(device=self.device,
                                        node=ZwavePlugin.node_to_dict(event.args['node']))
        elif event.signal == network.SIGNAL_CONTROLLER_COMMAND:
            event = ZwaveCommandEvent(device=self.device,
                                      state=event.args['state'],
                                      state_description=event.args['state_full'],
                                      error=event.args['error'] if event.args['error_int'] else None,
                                      error_description=event.args['error_full'] if event.args['error_int'] else None,
                                      node=ZwavePlugin.node_to_dict(event.args['node']) if event.args['node'] else None)
        elif event.signal == network.SIGNAL_CONTROLLER_WAITING:
            event = ZwaveCommandWaitingEvent(device=self.device,
                                             state=event.args['state'],
                                             state_description=event.args['state_full'])
        elif event.signal == network.SIGNAL_CREATE_BUTTON:
            event = ZwaveButtonCreatedEvent(device=self.device,
                                            node=ZwavePlugin.node_to_dict(event.args['node']))
        elif event.signal == network.SIGNAL_DELETE_BUTTON:
            event = ZwaveButtonRemovedEvent(device=self.device,
                                            node=ZwavePlugin.node_to_dict(event.args['node']))
        elif event.signal == network.SIGNAL_GROUP:
            event = ZwaveNodeGroupEvent(device=self.device,
                                        node=ZwavePlugin.node_to_dict(event.args['node']),
                                        group_index=event.args['groupidx'])
        elif event.signal == network.SIGNAL_NETWORK_AWAKED:
            event = ZwaveNetworkReadyEvent(
                device=self.device,
                ozw_library_version=self.network.controller.ozw_library_version,
                python_library_version=self.network.controller.python_library_version,
                zwave_library=self.network.controller.library_description,
                home_id=self.network.controller.home_id,
                node_id=self.network.controller.node_id,
                node_version=self.network.controller.node.version,
                nodes_count=self.network.nodes_count,
            )
        elif event.signal == network.SIGNAL_NODE_EVENT:
            event = ZwaveNodeEvent(device=self.device,
                                   node=ZwavePlugin.node_to_dict(event.args['node']))
        elif event.signal == network.SIGNAL_NODE_ADDED:
            event = ZwaveNodeAddedEvent(device=self.device,
                                        node=ZwavePlugin.node_to_dict(event.args['node']))
        elif event.signal == network.SIGNAL_NODE_NAMING:
            event = ZwaveNodeRenamedEvent(device=self.device,
                                          node=ZwavePlugin.node_to_dict(event.args['node']))
        elif event.signal == network.SIGNAL_NODE_READY:
            event = ZwaveNodeReadyEvent(device=self.device,
                                        node=ZwavePlugin.node_to_dict(event.args['node']))
        elif event.signal == network.SIGNAL_NODE_REMOVED:
            event = ZwaveNodeRemovedEvent(device=self.device,
                                          node=ZwavePlugin.node_to_dict(event.args['node']))
        elif event.signal == network.SIGNAL_POLLING_DISABLED:
            event = ZwaveNodePollingEnabledEvent(device=self.device,
                                                 node=ZwavePlugin.node_to_dict(event.args['node']))
        elif event.signal == network.SIGNAL_POLLING_ENABLED:
            event = ZwaveNodePollingDisabledEvent(device=self.device,
                                                  node=ZwavePlugin.node_to_dict(event.args['node']))
        elif event.signal == network.SIGNAL_SCENE_EVENT:
            event = ZwaveNodeSceneEvent(device=self.device,
                                        scene_id=event.args['scene_id'],
                                        node=ZwavePlugin.node_to_dict(event.args['node']))
        elif event.signal == network.SIGNAL_VALUE_ADDED:
            event = ZwaveValueAddedEvent(device=self.device,
                                         node=ZwavePlugin.node_to_dict(event.args['node']),
                                         value=ZwavePlugin.value_to_dict(event.args['value']))
        elif event.signal == network.SIGNAL_VALUE_CHANGED:
            event = ZwaveValueChangedEvent(device=self.device,
                                           node=ZwavePlugin.node_to_dict(event.args['node']),
                                           value=ZwavePlugin.value_to_dict(event.args['value']))
        elif event.signal == network.SIGNAL_VALUE_REFRESHED:
            event = ZwaveValueRefreshedEvent(device=self.device,
                                             node=ZwavePlugin.node_to_dict(event.args['node']),
                                             value=ZwavePlugin.value_to_dict(event.args['value']))
        elif event.signal == network.SIGNAL_VALUE_REMOVED:
            event = ZwaveValueRemovedEvent(device=self.device,
                                           node=ZwavePlugin.node_to_dict(event.args['node']),
                                           value=ZwavePlugin.value_to_dict(event.args['value']))
        else:
            self.logger.info('Received unhandled ZWave event: {}'.format(event))

        if isinstance(event, ZwaveEvent):
            self.bus.post(event)

    def __enter__(self):
        self.start_network()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.stop_network()

    def loop(self):
        try:
            event = event_queue.get(block=True, timeout=1.0)
            self._process_event(event)
        except queue.Empty:
            pass


# vim:sw=4:ts=4:et:



from typing import Any, Dict, Optional, List, Union

from openzwave.group import ZWaveGroup
from openzwave.node import ZWaveNode
from openzwave.scene import ZWaveScene
from openzwave.value import ZWaveValue

from platypush.backend.zwave import ZwaveBackend
from platypush.context import get_backend
from platypush.plugins import Plugin, action


class ZwavePlugin(Plugin):
    """
    This plugin interacts with the devices on a Z-Wave network started through the
    :class:`platypush.backend.zwave.ZwaveBackend` backend.

    Requires:

        * **python-openzwave** (``pip install python-openzwave``)
        * The :class:`platypush.backend.zwave.ZwaveBackend` backend configured and running.

    """

    @staticmethod
    def _get_backend() -> ZwaveBackend:
        backend = get_backend('zwave')
        if not backend:
            raise AssertionError('Z-Wave backend not configured')

        return backend

    @classmethod
    def _get_network(cls):
        backend = cls._get_backend()
        if not backend.network:
            backend.start_network()

        return backend.network

    @classmethod
    def _get_controller(cls):
        return cls._get_network().controller

    @action
    def start_network(self):
        backend = self._get_backend()
        backend.start_network()

    @action
    def stop_network(self):
        backend = self._get_backend()
        backend.stop_network()

    @action
    def status(self) -> Dict[str, Any]:
        """
        Get the status of the controller.
        :return: dict
        """
        backend = self._get_backend()
        network = self._get_network()
        controller = self._get_controller()

        return {
            'device': backend.device,
            'state': network.state_str,
            'stats': controller.stats,
        }

    @action
    def add_node(self, do_security=False):
        """
        Start the inclusion process to add a node to the network.

        :param do_security: Whether to initialize the Network Key on the device if it supports the Security CC
        """
        controller = self._get_controller()
        controller.add_node(do_security)

    @action
    def remove_node(self):
        """
        Remove a node from the network.
        """
        controller = self._get_controller()
        controller.remove_node()
        self.write_config()

    @action
    def remove_failed_node(self, node_id: Optional[int] = None, node_name: Optional[str] = None):
        """
        Remove a failed node from the network.

        :param node_id: Filter by node_id.
        :param node_name: Filter by node name.
        """
        controller = self._get_controller()
        node = self._get_node(node_id=node_id, node_name=node_name)
        controller.remove_failed_node(node.node_id)
        self.write_config()

    @action
    def replace_failed_node(self, node_id: Optional[int] = None, node_name: Optional[str] = None):
        """
        Replace a failed node on the network.

        :param node_id: Filter by node_id.
        :param node_name: Filter by node name.
        """
        controller = self._get_controller()
        node = self._get_node(node_id=node_id, node_name=node_name)
        controller.replace_failed_node(node.node_id)
        self.write_config()

    @action
    def replication_send(self, node_id: Optional[int] = None, node_name: Optional[str] = None):
        """
        Send node information from the primary to the secondary controller.

        :param node_id: Filter by node_id.
        :param node_name: Filter by node name.
        """
        controller = self._get_controller()
        node = self._get_node(node_id=node_id, node_name=node_name)
        controller.replication_send(node.node_id)

    @action
    def request_network_update(self, node_id: Optional[int] = None, node_name: Optional[str] = None):
        """
        Request a network update to a node.

        :param node_id: Filter by node_id.
        :param node_name: Filter by node name.
        """
        controller = self._get_controller()
        node = self._get_node(node_id=node_id, node_name=node_name)
        controller.request_network_update(node.node_id)

    @action
    def request_node_neighbour_update(self, node_id: Optional[int] = None, node_name: Optional[str] = None):
        """
        Request a neighbours list update to a node.

        :param node_id: Filter by node_id.
        :param node_name: Filter by node name.
        """
        controller = self._get_controller()
        node = self._get_node(node_id=node_id, node_name=node_name)
        controller.request_node_neighbor_update(node.node_id)

    @staticmethod
    def value_to_dict(value: Optional[ZWaveValue]) -> Dict[str, Any]:
        if not value:
            return {}

        return {
            'command_class': value.node.get_command_class_as_string(value.command_class),
            'data': value.data,
            'data_as_string': value.data_as_string,
            'data_items': list(value.data_items) if isinstance(value.data_items, set) else value.data_items,
            'genre': value.genre,
            'help': value.help,
            'home_id': value.home_id,
            'id_on_network': value.id_on_network if value.parent_id is not None and value.command_class is not None
            and value.instance is not None and value.index is not None else None,
            'index': value.index,
            'instance': value.instance,
            'is_polled': value.is_polled,
            'is_read_only': value.is_read_only,
            'is_set': value.is_set,
            'is_write_only': value.is_write_only,
            'label': value.label,
            'last_update': value.last_update,
            'min': value.min,
            'max': value.max,
            'object_id': value.object_id,
            'outdated': value.outdated,
            'parent_id': value.parent_id,
            'poll_intensity': value.poll_intensity,
            'precision': value.precision,
            'type': value.type,
            'units': value.units,
            'use_cache': value.use_cache,
            'value_id': value.value_id,
        }

    @staticmethod
    def group_to_dict(group: Optional[ZWaveGroup]) -> Dict[str, Any]:
        if not group:
            return {}

        return {
            'index': group.index,
            'label': group.label,
            'max_associations': group.max_associations,
            'associations': group.associations,
        }

    @classmethod
    def node_to_dict(cls, node: Optional[ZWaveNode]) -> Dict[str, Any]:
        if not node:
            return {}

        return {
            'node_id': node.node_id,
            'home_id': node.home_id,
            'capabilities': list(node.capabilities),
            'command_classes': [node.get_command_class_as_string(cc) for cc in node.command_classes],
            'device_type': node.device_type,
            'groups': {
                group_id: cls.group_to_dict(group)
                for group_id, group in node.groups.items()
            },
            'is_awake': node.is_awake,
            'is_failed': node.is_failed,
            'is_beaming_device': node.is_beaming_device,
            'is_frequent_listening_device': node.is_frequent_listening_device,
            'is_info_received': node.is_info_received,
            'is_listening_device': node.is_listening_device,
            'is_locked': node.is_locked,
            'is_ready': node.is_ready,
            'is_routing_device': node.is_routing_device,
            'is_security_device': node.is_security_device,
            'is_sleeping': node.is_sleeping,
            'last_update': node.last_update,
            'location': node.location,
            'manufacturer_id': node.manufacturer_id,
            'manufacturer_name': node.manufacturer_name,
            'max_baud_rate': node.max_baud_rate,
            'neighbours': [node_id for node_id in node.neighbors],
            'name': node.name,
            'outdated': node.outdated,
            'product_id': node.product_id,
            'product_name': node.product_name,
            'product_type': node.product_type,
            'query_stage': node.query_stage,
            'role': node.role,
            'security': node.security,
            'specific': node.specific,
            'type': node.type,
            'use_cache': node.use_cache,
            'version': node.version,
            'values': {
                value.id_on_network: cls.value_to_dict(value)
                for value_id, value in (node.values or {}).items()
            },
        }

    def _get_node(self, node_id: Optional[int] = None, node_name: Optional[str] = None) -> ZWaveNode:
        assert node_id is not None or node_name is not None, 'Specify either node_id or name'
        nodes = self._get_backend().network.nodes

        if node_id is not None:
            assert node_id in nodes, 'No such node_id: {}'.format(node_id)
            return nodes[node_id]

        nodes = [n for n in nodes.values() if n.name == node_name]
        assert nodes, 'No such node name: {}'.format(node_name)
        return nodes[0]

    def _get_groups(self) -> Dict[int, ZWaveGroup]:
        return {
            group_index: group
            for node in self._get_backend().network.nodes.values()
            for group_index, group in node.groups.items()
        }

    def _get_group(self, group_index: Optional[int] = None, group_label: Optional[str] = None) -> ZWaveGroup:
        assert group_index is not None or group_label is not None, 'Specify either group_index or label'
        groups = self._get_groups()

        if group_index is not None:
            assert group_index in groups, 'No such group_index: {}'.format(group_index)
            return groups[group_index]

        groups = [g for g in groups.values() if g.label == group_label]
        assert groups, 'No such group label: {}'.format(group_label)
        return groups[0]

    def _get_scene(self, scene_id: Optional[int] = None, scene_label: Optional[str] = None) -> ZWaveScene:
        assert scene_id is not None or scene_label is not None, 'Specify either scene_id or label'
        scenes = self._get_backend().network.get_scenes()

        if scene_id is not None:
            assert scene_id in scenes, 'No such scene_id: {}'.format(scene_id)
            return scenes[scene_id]

        scenes = [s for s in scenes.values() if s['label'] == scene_label]
        assert scenes, 'No such scene label: {}'.format(scene_label)
        return scenes[0]

    @action
    def get_nodes(self, node_id: Optional[int] = None, node_name: Optional[str] = None) -> Dict[str, Any]:
        """
        Get the nodes associated to the network.

        :param node_id: Filter by node_id.
        :param node_name: Filter by node name.
        """
        if node_id is not None or node_name is not None:
            return self.node_to_dict(self._get_node(node_id=node_id, node_name=node_name))

        return {
            node_id: self.node_to_dict(node)
            for node_id, node in self._get_backend().network.nodes.items()
        }

    @action
    def get_node_stats(self, node_id: Optional[int] = None, node_name: Optional[str] = None) -> Dict[str, Any]:
        """
        Get the statistics of a node on the network.

        :param node_id: Filter by node_id.
        :param node_name: Filter by node name.
        """
        node = self._get_node(node_id=node_id, node_name=node_name)
        return node.stats

    @action
    def set_node_name(self, new_name: str, node_id: Optional[int] = None, node_name: Optional[str] = None):
        """
        Rename a node on the network.

        :param new_name: New name for the node.
        :param node_id: Filter by node_id.
        :param node_name: Filter by current node name.
        """
        node = self._get_node(node_id=node_id, node_name=node_name)
        node.name = new_name
        self.write_config()

    @action
    def set_node_product_name(self, product_name: str, node_id: Optional[int] = None, node_name: Optional[str] = None):
        """
        Set the product name of a node.

        :param product_name: Product name.
        :param node_id: Filter by node_id.
        :param node_name: Filter by current node name.
        """
        node = self._get_node(node_id=node_id, node_name=node_name)
        node.product_name = product_name
        self.write_config()

    @action
    def set_node_manufacturer_name(self, manufacturer_name: str, node_id: Optional[int] = None,
                                   node_name: Optional[str] = None):
        """
        Set the manufacturer name of a node.

        :param manufacturer_name: Manufacturer name.
        :param node_id: Filter by node_id.
        :param node_name: Filter by current node name.
        """
        node = self._get_node(node_id=node_id, node_name=node_name)
        node.manufacturer_name = manufacturer_name
        self.write_config()

    @action
    def set_node_location(self, location: str, node_id: Optional[int] = None, node_name: Optional[str] = None):
        """
        Set the location of a node.

        :param location: Node location.
        :param node_id: Filter by node_id.
        :param node_name: Filter by current node name.
        """
        node = self._get_node(node_id=node_id, node_name=node_name)
        node.location = location
        self.write_config()

    @action
    def cancel_command(self):
        """
        Cancel the current running command.
        """
        self._get_controller().cancel_command()

    @action
    def kill_command(self):
        """
        Immediately terminate any running command on the controller and release the lock.
        """
        self._get_controller().kill_command()

    @action
    def set_controller_name(self, name: str):
        """
        Set the name of the controller on the network.

        :param name: New controller name.
        """
        self._get_controller().name = name
        self.write_config()

    @action
    def get_capabilities(self) -> List[str]:
        """
        Get the capabilities of the controller.
        """
        return list(self._get_controller().capabilities)

    @action
    def receive_configuration(self):
        """
        Receive the configuration from the primary controller on the network. Requires a primary controller active.
        """
        self._get_controller().receive_configuration()

    @action
    def transfer_primary_role(self):
        """
        Add a new controller to the network and make it the primary.
        The existing primary will become a secondary controller.
        """
        self._get_controller().transfer_primary_role()

    @action
    def heal(self, refresh_routes: bool = False):
        """
        Heal network by requesting nodes rediscover their neighbors.

        :param refresh_routes: Whether to perform return routes initialization (default: ``False``).
        """
        self._get_network().heal(refresh_routes)

    @action
    def switch_all(self, state: bool):
        """
        Switch all the connected devices on/off.

        :param state: True (switch on) or False (switch off).
        """
        self._get_network().switch_all(state)

    @action
    def test(self, count: int = 1):
        """
        Send a number of test messages to every node and record results.

        :param count: The number of test messages to send.
        """
        self._get_network().test(count)

    def _get_value(self, value_id: Optional[int] = None, id_on_network: Optional[str] = None,
                   node_id: Optional[int] = None, node_name: Optional[str] = None, value_label: Optional[str] = None) \
            -> ZWaveValue:
        assert (value_id is not None or id_on_network is not None) or \
               ((node_id is not None or node_name is not None) and value_label is not None),\
            'Specify either value_id, id_on_network, or [node_id/node_name, value_label]'

        if value_id is not None:
            return self._get_network().get_value(value_id)
        if id_on_network is not None:
            values = [value
                      for node in self._get_network().nodes.values()
                      for value in node.values.values()
                      if value.id_on_network == id_on_network]

            assert values, 'No such value ID: {}'.format(id_on_network)
            return values[0]

        node = self._get_node(node_id=node_id, node_name=node_name)
        values = [v for v in node.values.values() if v.label == value_label]
        assert values, 'No such value on node "{}": "{}"'.format(node.name, value_label)
        return values[0]

    @action
    def get_value(self, value_id: Optional[int] = None, id_on_network: Optional[str] = None,
                  value_label: Optional[str] = None, node_id: Optional[int] = None, node_name: Optional[str] = None) \
            -> Dict[str, Any]:
        """
        Get a value on the network.

        :param value_id: Select by value_id.
        :param id_on_network: Select value by id_on_network.
        :param value_label: Select value by [node_id/node_name, value_label]
        :param node_id: Select value by [node_id/node_name, value_label]
        :param node_name: Select value by [node_id/node_name, value_label]
        """
        return self._get_value(value_id=value_id, id_on_network=id_on_network,
                               node_id=node_id, node_name=node_name, value_label=value_label).to_dict()

    @action
    def set_value(self, data, value_id: Optional[int] = None, id_on_network: Optional[str] = None,
                  value_label: Optional[str] = None, node_id: Optional[int] = None, node_name: Optional[str] = None):
        """
        Set a value.

        :param data: Data to set for the value.
        :param value_id: Select value by value_id.
        :param id_on_network: Select value by id_on_network.
        :param value_label: Select value by [node_id/node_name, value_label]
        :param node_id: Select value by [node_id/node_name, value_label]
        :param node_name: Select value by [node_id/node_name, value_label]
        """
        value = self._get_value(value_id=value_id, id_on_network=id_on_network,
                                node_id=node_id, node_name=node_name, value_label=value_label)
        new_val = value.check_data(data)
        assert new_val is not None, 'Invalid value passed to the property'
        node: ZWaveNode = self._get_network().nodes[value.node.node_id]
        node.values[value.value_id].data = new_val
        self.write_config()

    @action
    def set_value_label(self, new_label: str, value_id: Optional[int] = None, id_on_network: Optional[str] = None,
                        value_label: Optional[str] = None, node_id: Optional[int] = None,
                        node_name: Optional[str] = None):
        """
        Change the label/name of a value.

        :param new_label: New value label.
        :param value_id: Select value by value_id.
        :param id_on_network: Select value by id_on_network.
        :param value_label: Select value by [node_id/node_name, value_label]
        :param node_id: Select value by [node_id/node_name, value_label]
        :param node_name: Select value by [node_id/node_name, value_label]
        """
        value = self._get_value(value_id=value_id, id_on_network=id_on_network,
                                node_id=node_id, node_name=node_name, value_label=value_label)
        value.label = new_label
        self.write_config()

    @action
    def node_add_value(self, value_id: Optional[int] = None, id_on_network: Optional[str] = None,
                       value_label: Optional[str] = None, node_id: Optional[int] = None,
                       node_name: Optional[str] = None):
        """
        Add a value to a node.

        :param value_id: Select value by value_id.
        :param id_on_network: Select value by id_on_network.
        :param value_label: Select value by label.
        :param node_id: Select node by node_id.
        :param node_name: Select node by label.
        """
        node = self._get_node(node_id=node_id, node_name=node_name)
        value = self._get_value(value_id=value_id, id_on_network=id_on_network, node_id=node.node_id,
                                value_label=value_label)
        node.add_value(value.value_id)
        self.write_config()

    @action
    def node_remove_value(self, value_id: Optional[int] = None, id_on_network: Optional[str] = None,
                          value_label: Optional[str] = None, node_id: Optional[int] = None,
                          node_name: Optional[str] = None):
        """
        Remove a value from a node.

        :param value_id: Select value by value_id.
        :param id_on_network: Select value by id_on_network.
        :param value_label: Select value by [node_id/node_name, value_label]
        :param node_id: Select node by node_id.
        :param node_name: Select node by label.
        """
        node = self._get_node(node_id=node_id, node_name=node_name)
        value = self._get_value(value_id=value_id, id_on_network=id_on_network, node_id=node.node_id,
                                value_label=value_label)
        node.remove_value(value.value_id)
        self.write_config()

    @action
    def node_heal(self, node_id: Optional[int] = None, node_name: Optional[str] = None, refresh_routes: bool = False):
        """
        Heal network node by requesting the node to rediscover their neighbours.

        :param node_id: Select node by node_id.
        :param node_name: Select node by label.
        :param refresh_routes: Whether to perform return routes initialization. (default: ``False``).
        """
        node = self._get_node(node_id=node_id, node_name=node_name)
        node.heal(refresh_routes)

    @action
    def node_update_neighbours(self, node_id: Optional[int] = None, node_name: Optional[str] = None):
        """
        Ask a node to update its neighbours table.

        :param node_id: Select node by node_id.
        :param node_name: Select node by label.
        """
        node = self._get_node(node_id=node_id, node_name=node_name)
        node.neighbor_update()

    @action
    def node_network_update(self, node_id: Optional[int] = None, node_name: Optional[str] = None):
        """
        Update the controller with network information.

        :param node_id: Select node by node_id.
        :param node_name: Select node by label.
        """
        node = self._get_node(node_id=node_id, node_name=node_name)
        node.network_update()

    @action
    def node_refresh_info(self, node_id: Optional[int] = None, node_name: Optional[str] = None):
        """
        Fetch up-to-date information about the node.

        :param node_id: Select node by node_id.
        :param node_name: Select node by label.
        """
        node = self._get_node(node_id=node_id, node_name=node_name)
        node.refresh_info()

    def _get_values(self, item: str, node_id: Optional[int] = None, node_name: Optional[str] = None) -> Dict[int, Any]:
        nodes = [self._get_node(node_id=node_id, node_name=node_name)] if node_id or node_name \
            else self._get_network().nodes.values()

        return {
            value.id_on_network: {
                'node_id': node.node_id,
                'node_name': node.name,
                **self.value_to_dict(value)
            }
            for node in nodes
            for value_id, value in getattr(node, 'get_' + item)().items()
        }

    @action
    def get_dimmers(self, node_id: Optional[int] = None, node_name: Optional[str] = None) -> Dict[int, Any]:
        """
        Get the dimmers on the network or associated to a node.

        :param node_id: Select node by node_id.
        :param node_name: Select node by label.
        """
        return self._get_values('dimmers', node_id=node_id, node_name=node_name)

    @action
    def get_node_config(self, node_id: Optional[int] = None, node_name: Optional[str] = None) -> Dict[int, Any]:
        """
        Get the configuration values of a node or of all the nodes on the network.

        :param node_id: Select node by node_id.
        :param node_name: Select node by label.
        """
        return self._get_values('configs', node_id=node_id, node_name=node_name)

    @action
    def get_battery_levels(self, node_id: Optional[int] = None, node_name: Optional[str] = None) -> Dict[int, Any]:
        """
        Get the battery levels of a node or of all the nodes on the network.

        :param node_id: Select node by node_id.
        :param node_name: Select node by name.
        """
        return self._get_values('battery_levels', node_id=node_id, node_name=node_name)

    @action
    def get_power_levels(self, node_id: Optional[int] = None, node_name: Optional[str] = None) -> Dict[int, Any]:
        """
        Get the power levels of this node.

        :param node_id: Select node by node_id.
        :param node_name: Select node by name.
        """
        return self._get_values('power_levels', node_id=node_id, node_name=node_name)

    @action
    def get_bulbs(self, node_id: Optional[int] = None, node_name: Optional[str] = None) -> Dict[int, Any]:
        """
        Get the bulbs/LEDs on the network or associated to a node.

        :param node_id: Select node by node_id.
        :param node_name: Select node by name.
        """
        return self._get_values('rgbbulbs', node_id=node_id, node_name=node_name)

    @action
    def get_switches(self, node_id: Optional[int] = None, node_name: Optional[str] = None) -> Dict[int, Any]:
        """
        Get the switches on the network or associated to a node.

        :param node_id: Select node by node_id.
        :param node_name: Select node by name.
        """
        return self._get_values('switches', node_id=node_id, node_name=node_name)

    @action
    def get_sensors(self, node_id: Optional[int] = None, node_name: Optional[str] = None) -> Dict[int, Any]:
        """
        Get the sensors on the network or associated to a node.

        :param node_id: Select node by node_id.
        :param node_name: Select node by name.
        """
        return self._get_values('sensors', node_id=node_id, node_name=node_name)

    @action
    def get_doorlocks(self, node_id: Optional[int] = None, node_name: Optional[str] = None) -> Dict[int, Any]:
        """
        Get the doorlocks on the network or associated to a node.

        :param node_id: Select node by node_id.
        :param node_name: Select node by name.
        """
        return self._get_values('doorlocks', node_id=node_id, node_name=node_name)

    @action
    def get_usercodes(self, node_id: Optional[int] = None, node_name: Optional[str] = None) -> Dict[int, Any]:
        """
        Get the usercodes on the network or associated to a node.

        :param node_id: Select node by node_id.
        :param node_name: Select node by name.
        """
        return self._get_values('usercodes', node_id=node_id, node_name=node_name)

    @action
    def get_thermostats(self, node_id: Optional[int] = None, node_name: Optional[str] = None) -> Dict[int, Any]:
        """
        Get the thermostats on the network or associated to a node.

        :param node_id: Select node by node_id.
        :param node_name: Select node by name.
        """
        return self._get_values('thermostats', node_id=node_id, node_name=node_name)

    @action
    def get_protections(self, node_id: Optional[int] = None, node_name: Optional[str] = None) -> Dict[int, Any]:
        """
        Get the protection-compatible devices on the network or associated to a node.

        :param node_id: Select node by node_id.
        :param node_name: Select node by name.
        """
        return self._get_values('protections', node_id=node_id, node_name=node_name)

    @action
    def get_groups(self) -> Dict[int, Any]:
        """
        Get the groups on the network.
        """
        return {
            group_index: self.group_to_dict(group)
            for group_index, group in self._get_groups().items()
        }

    @action
    def get_scenes(self) -> Dict[str, Any]:
        """
        Get the scenes configured on the network.
        """
        network = self._get_network()
        if not network.get_scenes():
            return {}

        return network.scenes_to_dict()

    @action
    def create_scene(self, label: str):
        """
        Create a new scene.

        :param label: Scene label.
        """
        self._get_network().create_scene(label)
        self.write_config()

    @action
    def remove_scene(self, scene_id: Optional[int] = None, scene_label: Optional[str] = None):
        """
        Remove a scene.

        :param scene_id: Select by scene_id.
        :param scene_label: Select by scene label.
        """
        scene = self._get_scene(scene_id=scene_id, scene_label=scene_label)
        self._get_network().remove_scene(scene.scene_id)
        self.write_config()

    @action
    def activate_scene(self, scene_id: Optional[int] = None, scene_label: Optional[str] = None):
        """
        Activate a scene.

        :param scene_id: Select by scene_id.
        :param scene_label: Select by scene label.
        """
        scene = self._get_scene(scene_id=scene_id, scene_label=scene_label)
        scene.activate()

    @action
    def set_scene_label(self, new_label: str, scene_id: Optional[int] = None, scene_label: Optional[str] = None):
        """
        Rename a scene/set the scene label.

        :param new_label: New label.
        :param scene_id: Select by scene_id.
        :param scene_label: Select by current scene label.
        """
        scene = self._get_scene(scene_id=scene_id, scene_label=scene_label)
        scene.label = new_label
        self.write_config()

    @action
    def scene_add_value(self, data: Optional[Any] = None,
                        value_id: Optional[int] = None, id_on_network: Optional[str] = None,
                        value_label: Optional[str] = None, scene_id: Optional[int] = None,
                        scene_label: Optional[str] = None, node_id: Optional[int] = None,
                        node_name: Optional[str] = None):
        """
        Add a value to a scene.

        WARNING: This method actually doesn't work, by own admission of the
        :ref:`OpenZWave developer <https://github.com/OpenZWave/python-openzwave/blob/master/src-lib/libopenzwave/libopenzwave.pyx#L4730>`_


        :param data: Data to set for the value (default: current value data).
        :param value_id: Select value by value_id.
        :param id_on_network: Select value by id_on_network.
        :param value_label: Select value by [node_id/node_name, value_label]
        :param node_id: Select value by [node_id/node_name, value_label]
        :param node_name: Select value by [node_id/node_name, value_label]
        :param scene_id: Select scene by scene_id.
        :param scene_label: Select scene by scene label.
        """
        value = self._get_value(value_id=value_id, id_on_network=id_on_network, node_id=node_id, node_name=node_name,
                                value_label=value_label)
        scene = self._get_scene(scene_id=scene_id, scene_label=scene_label)
        data = data if data is not None else value.data
        data = value.check_data(data)
        assert data is not None, 'Invalid value passed to the property'

        scene.add_value(value.value_id, data)
        self.write_config()

    @action
    def scene_remove_value(self, value_id: Optional[int] = None, id_on_network: Optional[str] = None,
                           value_label: Optional[str] = None, scene_id: Optional[int] = None,
                           scene_label: Optional[str] = None, node_id: Optional[int] = None,
                           node_name: Optional[str] = None):
        """
        Remove a value from a scene.

        :param value_id: Select value by value_id.
        :param id_on_network: Select value by id_on_network.
        :param value_label: Select value by [node_id/node_name, value_label]
        :param node_id: Select value by [node_id/node_name, value_label]
        :param node_name: Select value by [node_id/node_name, value_label]
        :param scene_id: Select scene by scene_id.
        :param scene_label: Select scene by scene label.
        """
        value = self._get_value(value_id=value_id, id_on_network=id_on_network, node_id=node_id, node_name=node_name,
                                value_label=value_label)
        scene = self._get_scene(scene_id=scene_id, scene_label=scene_label)
        scene.remove_value(value.value_id)
        self.write_config()

    @action
    def get_scene_values(self, scene_id: Optional[int] = None, scene_label: Optional[str] = None) -> dict:
        """
        Get the values associated to a scene.

        :param scene_id: Select by scene_id.
        :param scene_label: Select by scene label.
        :return: value_id -> value (as a dict) mapping.
        """
        scene = self._get_scene(scene_id=scene_id, scene_label=scene_label)
        return {v.value_id: v.to_dict() for v in scene.get_values().items()}

    @action
    def get_scene_values(self, scene_id: Optional[int] = None, scene_label: Optional[str] = None) -> dict:
        """
        Get the values associated to a scene.

        :param scene_id: Select by scene_id.
        :param scene_label: Select by scene label.
        """
        scene = self._get_scene(scene_id=scene_id, scene_label=scene_label)
        return {v.value_id: v.to_dict() for v in scene.get_values().items()}

    @action
    def create_button(self, button_id: Union[int, str], node_id: Optional[int] = None, node_name: Optional[str] = None):
        """
        Create a handheld button on a device. Only intended for bridge firmware controllers.

        :param button_id: The ID of the button.
        :param node_id: Filter by node_id.
        :param node_name: Filter by current node name.
        """
        node = self._get_node(node_id=node_id, node_name=node_name)
        self._get_controller().create_button(node.node_id, button_id)
        self.write_config()

    @action
    def delete_button(self, button_id: Union[int, str], node_id: Optional[int] = None, node_name: Optional[str] = None):
        """
        Delete a button association from a device. Only intended for bridge firmware controllers.

        :param button_id: The ID of the button.
        :param node_id: Filter by node_id.
        :param node_name: Filter by current node name.
        """
        node = self._get_node(node_id=node_id, node_name=node_name)
        self._get_controller().delete_button(node.node_id, button_id)
        self.write_config()

    @action
    def add_node_to_group(self, group_index: Optional[int] = None, group_label: Optional[str] = None,
                          node_id: Optional[int] = None, node_name: Optional[str] = None):
        """
        Add a node to a group.

        :param group_index: Select group by group index.
        :param group_label: Select group by group label.
        :param node_id: Select node by node_id.
        :param node_name: Select node by node name.
        :return:
        """
        node = self._get_node(node_id=node_id, node_name=node_name)
        group = self._get_group(group_index=group_index, group_label=group_label)
        group.add_association(node.node_id)
        self.write_config()

    @action
    def remove_node_from_group(self, group_index: Optional[int] = None, group_label: Optional[str] = None,
                               node_id: Optional[int] = None, node_name: Optional[str] = None):
        """
        Remove a node from a group.

        :param group_index: Select group by group index.
        :param group_label: Select group by group label.
        :param node_id: Select node by node_id.
        :param node_name: Select node by node name.
        :return:
        """
        node = self._get_node(node_id=node_id, node_name=node_name)
        group = self._get_group(group_index=group_index, group_label=group_label)
        group.remove_association(node.node_id)
        self.write_config()

    @action
    def create_new_primary(self):
        """
        Create a new primary controller on the network when the previous primary fails.
        """
        self._get_controller().create_new_primary()
        self.write_config()

    @action
    def hard_reset(self):
        """
        Perform a hard reset of the controller. It erases its network configuration settings.
        The controller becomes a primary controller ready to add devices to a new network.
        """
        self._get_controller().hard_reset()

    @action
    def soft_reset(self):
        """
        Perform a soft reset of the controller.
        Resets a controller without erasing its network configuration settings.
        """
        self._get_controller().soft_reset()

    @action
    def write_config(self):
        """
        Store the current configuration of the network to the user directory.
        """
        self._get_network().write_config()


# vim:sw=4:ts=4:et:



import base64
import os
import re
import select

from typing import Dict, Optional

from platypush.plugins.sensor import SensorPlugin

from platypush.plugins import action
from platypush.message.response.bluetooth import BluetoothScanResponse, \
    BluetoothLookupNameResponse, BluetoothLookupServiceResponse, BluetoothResponse


class BluetoothPlugin(SensorPlugin):
    """
    Bluetooth plugin

    Requires:

        * **pybluez** (``pip install pybluez``)
        * **pyobex** (``pip install git+https://github.com/BlackLight/PyOBEX``)

    """

    import bluetooth

    class _DeviceDiscoverer(bluetooth.DeviceDiscoverer):
        def __init__(self, name, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self.name = name
            self.device = {}
            self.done = True

        def pre_inquiry(self):
            self.done = False

        def device_discovered(self, dev_addr, dev_class, rssi, dev_name):
            dev_name = dev_name.decode()
            if dev_name == self.name:
                self.device = {
                    'addr': dev_addr,
                    'name': dev_name,
                    'class': dev_class,
                }

                self.done = True

        def inquiry_complete(self):
            self.done = True

    def __init__(self, device_id: int = -1, **kwargs):
        """
        :param device_id: Default adapter device_id to be used (default: -1, auto)
        """
        super().__init__(**kwargs)
        self.device_id = device_id
        self._devices = []
        self._devices_by_addr = {}
        self._devices_by_name = {}
        self._port_and_protocol_by_addr_and_srv_uuid = {}
        self._port_and_protocol_by_addr_and_srv_name = {}
        self._socks = {}

    def _get_device_addr(self, device):
        if re.match('([0-9A-F]{2}:){5}[0-9A-F]{2}', device, re.IGNORECASE):
            return device
        if device in self._devices_by_name:
            return self._devices_by_name[device]['addr']

        return self.lookup_address(device).output['addr']

    @action
    def scan(self, device_id: Optional[int] = None, duration: int = 10) -> BluetoothScanResponse:
        """
        Scan for nearby bluetooth devices

        :param device_id: Bluetooth adapter ID to use (default configured if None)
        :param duration: Scan duration in seconds
        """
        from bluetooth import discover_devices

        if device_id is None:
            device_id = self.device_id

        self.logger.debug('Discovering devices on adapter {}, duration: {} seconds'.format(
            device_id, duration))

        devices = discover_devices(duration=duration, lookup_names=True, lookup_class=True, device_id=device_id,
                                   flush_cache=True)
        response = BluetoothScanResponse(devices)

        self._devices = response.devices
        self._devices_by_addr = {dev['addr']: dev for dev in self._devices}
        self._devices_by_name = {dev['name']: dev for dev in self._devices if dev.get('name')}
        return response

    @action
    def get_measurement(self, device_id: Optional[int] = None, duration: Optional[int] = 10, *args, **kwargs) \
            -> Dict[str, dict]:
        """
        Wrapper for ``scan`` that returns bluetooth devices in a format usable by sensor backends.

        :param device_id: Bluetooth adapter ID to use (default configured if None)
        :param duration: Scan duration in seconds
        :return: Device address -> info map.
        """
        devices = self.scan(device_id=device_id, duration=duration).output
        return {device['addr']: device for device in devices}

    @action
    def lookup_name(self, addr: str, timeout: int = 10) -> BluetoothLookupNameResponse:
        """
        Look up the name of a nearby bluetooth device given the address

        :param addr: Device address
        :param timeout: Lookup timeout (default: 10 seconds)
        """
        from bluetooth import lookup_name

        self.logger.info('Looking up name for device {}'.format(addr))
        name = lookup_name(addr, timeout=timeout)

        dev = {
            'addr': addr,
            'name': name,
            'class': self._devices_by_addr.get(addr, {}).get('class'),
        }

        self._devices_by_addr[addr] = dev
        if name:
            self._devices_by_name[name] = dev

        return BluetoothLookupNameResponse(addr=addr, name=name)

    @action
    def lookup_address(self, name: str, timeout: int = 10) -> BluetoothLookupNameResponse:
        """
        Look up the address of a nearby bluetooth device given the name

        :param name: Device name
        :param timeout: Lookup timeout (default: 10 seconds)
        """

        self.logger.info('Looking up address for device {}'.format(name))
        discoverer = self._DeviceDiscoverer(name)
        discoverer.find_devices(lookup_names=True, duration=timeout)
        readfiles = [discoverer]

        while True:
            rfds = select.select(readfiles, [], [])[0]
            if discoverer in rfds:
                discoverer.process_event()

            if discoverer.done:
                break

        dev = discoverer.device
        if not dev:
            raise RuntimeError('No such device: {}'.format(name))

        addr = dev.get('addr')
        self._devices_by_addr[addr] = dev
        self._devices_by_name[name] = dev
        return BluetoothLookupNameResponse(addr=addr, name=name)

    @action
    def find_service(self, name: str = None, addr: str = None, uuid: str = None) -> BluetoothLookupServiceResponse:
        """
        Look up for a service published by a nearby bluetooth device. If all the parameters are null then all the
        published services on the nearby devices will be returned. See
        `:class:platypush.message.response.bluetoothBluetoothLookupServiceResponse` for response structure reference.

        :param name: Service name
        :param addr: Service/device address
        :param uuid: Service UUID
        """

        import bluetooth
        from bluetooth import find_service
        services = find_service(name=name, address=addr, uuid=uuid)

        self._port_and_protocol_by_addr_and_srv_uuid.update({
            (srv['host'], srv['service-id']): (srv['port'], getattr(bluetooth, srv['protocol']))
            for srv in services if srv.get('service-id')
        })

        self._port_and_protocol_by_addr_and_srv_name.update({
            (srv['host'], srv['name']): (srv['port'], getattr(bluetooth, srv['protocol']))
            for srv in services if srv.get('name')
        })

        return BluetoothLookupServiceResponse(services)

    def _get_sock(self, protocol=None, device: str = None, port: int = None, service_uuid: str = None,
                  service_name: str = None, connect_if_closed=False):
        sock = None
        addr = self._get_device_addr(device)

        if not (addr and port and protocol):
            addr, port, protocol = self._get_addr_port_protocol(protocol=protocol, device=device, port=port,
                                                                service_uuid=service_uuid, service_name=service_name)

        if (addr, port) in self._socks:
            sock = self._socks[(addr, port)]
        elif connect_if_closed:
            self.connect(protocol=protocol, device=device, port=port, service_uuid=service_uuid,
                         service_name=service_name)
            sock = self._socks[(addr, port)]

        return sock

    def _get_addr_port_protocol(self, protocol=None, device: str = None, port: int = None, service_uuid: str = None,
                                service_name: str = None) -> tuple:
        import bluetooth

        addr = self._get_device_addr(device) if device else None
        if service_uuid or service_name:
            if addr:
                if service_uuid:
                    (port, protocol) = self._port_and_protocol_by_addr_and_srv_uuid[(addr, service_uuid)] \
                        if (addr, service_uuid) in self._port_and_protocol_by_addr_and_srv_uuid else \
                        (None, None)
                else:
                    (port, protocol) = self._port_and_protocol_by_addr_and_srv_name[(addr, service_name)] \
                        if (addr, service_name) in self._port_and_protocol_by_addr_and_srv_name else \
                        (None, None)

            if not (addr and port):
                self.logger.info('Discovering devices, service_name={name}, uuid={uuid}, address={addr}'.format(
                    name=service_name, uuid=service_uuid, addr=addr))

                services = [
                    srv for srv in self.find_service().services
                    if (service_name is None or srv.get('name') == service_name) and
                       (addr is None or srv.get('host') == addr) and
                       (service_uuid is None or srv.get('service-id') == service_uuid)
                ]

                if not services:
                    raise RuntimeError('No such service: name={name} uuid={uuid} address={addr}'.format(
                        name=service_name, uuid=service_uuid, addr=addr))

                service = services[0]
                addr = service['host']
                port = service['port']
                protocol = getattr(bluetooth, service['protocol'])
        elif protocol:
            if isinstance(protocol, str):
                protocol = getattr(bluetooth, protocol)
        else:
            raise RuntimeError('No service name/UUID nor bluetooth protocol (RFCOMM/L2CAP) specified')

        if not (addr and port):
            raise RuntimeError('No valid device name/address, port, service name or UUID specified')

        return addr, port, protocol

    @action
    def connect(self, protocol=None, device: str = None, port: int = None, service_uuid: str = None,
                service_name: str = None):
        """
        Connect to a bluetooth device.
        You can query the advertised services through ``find_service``.

        :param protocol: Supported values: either 'RFCOMM'/'L2CAP' (str) or bluetooth.RFCOMM/bluetooth.L2CAP
            int constants (int)
        :param device: Device address or name
        :param port: Port number
        :param service_uuid: Service UUID
        :param service_name: Service name
        """
        from bluetooth import BluetoothSocket

        addr, port, protocol = self._get_addr_port_protocol(protocol=protocol, device=device, port=port,
                                                            service_uuid=service_uuid, service_name=service_name)
        sock = self._get_sock(protocol=protocol, device=addr, port=port)
        if sock:
            self.close(device=addr, port=port)

        sock = BluetoothSocket(protocol)
        self.logger.info('Opening connection to device {} on port {}'.format(addr, port))
        sock.connect((addr, port))
        self.logger.info('Connected to device {} on port {}'.format(addr, port))
        self._socks[(addr, port)] = sock

    @action
    def close(self, device: str = None, port: int = None, service_uuid: str = None, service_name: str = None):
        """
        Close an active bluetooth connection

        :param device: Device address or name
        :param port: Port number
        :param service_uuid: Service UUID
        :param service_name: Service name
        """
        sock = self._get_sock(device=device, port=port, service_uuid=service_uuid, service_name=service_name)

        if not sock:
            self.logger.debug('Close on device {}({}) that is not connected'.format(device, port))
            return

        try:
            sock.close()
        except Exception as e:
            self.logger.warning('Exception while closing previous connection to {}({}): {}'.format(
                device, port, str(e)))

    @action
    def send(self, data, device: str = None, port: int = None, service_uuid: str = None, service_name: str = None,
             binary: bool = False):
        """
        Send data to an active bluetooth connection

        :param data: Data to be sent
        :param device: Device address or name
        :param service_uuid: Service UUID
        :param service_name: Service name
        :param port: Port number
        :param binary: Set to true if msg is a base64-encoded binary string
        """
        from bluetooth import BluetoothError

        sock = self._get_sock(device=device, port=port, service_uuid=service_uuid, service_name=service_name,
                              connect_if_closed=True)

        if binary:
            data = base64.decodebytes(data.encode() if isinstance(data, str) else data)

        try:
            sock.send(data)
        except BluetoothError as e:
            self.close(device=device, port=port, service_uuid=service_uuid, service_name=service_name)
            raise e

    @action
    def recv(self, device: str, port: int, service_uuid: str = None, service_name: str = None, size: int = 1024,
             binary: bool = False) -> BluetoothResponse:
        """
        Send data to an active bluetooth connection

        :param device: Device address or name
        :param port: Port number
        :param service_uuid: Service UUID
        :param service_name: Service name
        :param size: Maximum number of bytes to be read
        :param binary: Set to true to return a base64-encoded binary string
        """
        from bluetooth import BluetoothError

        sock = self._get_sock(device=device, port=port, service_uuid=service_uuid, service_name=service_name,
                              connect_if_closed=True)

        if not sock:
            self.connect(device=device, port=port, service_uuid=service_uuid, service_name=service_name)
            sock = self._get_sock(device=device, port=port, service_uuid=service_uuid, service_name=service_name)

        try:
            data = sock.recv(size)
        except BluetoothError as e:
            self.close(device=device, port=port, service_uuid=service_uuid, service_name=service_name)
            raise e

        if binary:
            data = base64.encodebytes(data)

        return BluetoothResponse(output=data.decode())

    @action
    def set_l2cap_mtu(self, mtu: int, device: str = None, port: int = None, service_name: str = None,
                      service_uuid: str = None):
        """
        Set the L2CAP MTU (Maximum Transmission Unit) value for a connected bluetooth device.
        Both the devices usually use the same MTU value over a connection.

        :param device: Device address or name
        :param port: Port number
        :param service_uuid: Service UUID
        :param service_name: Service name
        :param mtu: New MTU value
        """
        from bluetooth import BluetoothError, set_l2cap_mtu, L2CAP

        sock = self._get_sock(protocol=L2CAP, device=device, port=port, service_uuid=service_uuid,
                              service_name=service_name, connect_if_closed=True)

        if not sock:
            raise RuntimeError('set_l2cap_mtu: device not connected')

        try:
            set_l2cap_mtu(sock, mtu)
        except BluetoothError as e:
            self.close(device=device, port=port, service_name=service_name, service_uuid=service_uuid)
            raise e

    @action
    def send_file(self, filename: str, device: str, port: int = None, data=None, service_name='OBEX Object Push',
                  binary: bool = False):
        """
        Send a local file to a device that exposes an OBEX Object Push service

        :param filename: Path of the file to be sent
        :param data: Alternatively to a file on disk you can send raw (string or binary) content
        :param device: Device address or name
        :param port: Port number
        :param service_name: Service name
        :param binary: Set to true if data is a base64-encoded binary string
        """
        from PyOBEX.client import Client

        if not data:
            filename = os.path.abspath(os.path.expanduser(filename))
            with open(filename, 'r') as f:
                data = f.read()
            filename = os.path.basename(filename)
        else:
            if binary:
                data = base64.decodebytes(data.encode() if isinstance(data, str) else data)

        addr, port, protocol = self._get_addr_port_protocol(device=device, port=port,
                                                            service_name=service_name)

        client = Client(addr, port)
        self.logger.info('Connecting to device {}'.format(addr))
        client.connect()
        self.logger.info('Sending file {} to device {}'.format(filename, addr))
        client.put(filename, data)
        self.logger.info('File {} sent to device {}'.format(filename, addr))
        client.disconnect()


# vim:sw=4:ts=4:et:
